"""
Day 11 mod: Candy Cane
Daily effect, December 11 Modified to resemble a Candy Cane pattern

Red and white sweeping diagonally over the lights.

Contributed by gonzotek
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
ctr.set_mode("off")

ctr.adjust_layout_aspect(1.0)  # How many times wider than high is the led installation?
cols = [hsl_color(0.6226, 1.0, -0.5631), hsl_color(0.6226, 1.0, -0.5631), hsl_color(0.5830, 1.0000, 0.5121), hsl_color(0.6226, 1.0, -0.5631), hsl_color(0.6226, 1.0, -0.5631), hsl_color(0.5830, 1.0000, 0.5121), hsl_color(0.6226, 1.0, -0.5631), hsl_color(0.6226, 1.0, -0.5631), hsl_color(0.5830, 1.0000, 0.5121), hsl_color(0.6030, 1.0000, 0.2121)]
GradientSequence(ctr, cols, speed=0.05, folds=0.8, angle=25).launch_movie()



"""
Day 30 - Random walk in colorspace with random gradient direction

Another variant of the Day 25 effect. Here the colors form a gradient that
moves over the leds at a slowly and randomly changing angle and change rate.
This is again a continuous effect. Aspect ratio is important as usual, so adjust below.
However, dimensionality or density of layout should matter less.
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
ctr.adjust_layout_aspect(1.0)  # How many times wider than high is the led installation?
eff = MeanderingSequence(ctr)
oldmode = ctr.get_mode()["mode"]
eff.launch_rt()
print("Started continuous effect - press Return to stop it")
input()
eff.stop_rt()
ctr.set_mode(oldmode)



"""
Day 20 - Kaleidoscope

Slowly changing symmetric pattern of random colors.
Again, looks best on a 2D layout, and with correctly adjusted aspect:
Change the number in the call to adjust_layout_aspect in the next to
last line of the code below.
"""

from xled_plus.samples.sample_setup import *

class SoftKaleidoScene(MovingShapesScene):
    def __init__(self, ctr, sym, colfunc):
        super(SoftKaleidoScene, self).__init__(ctr)
        self.freq = 0.6
        self.symang = m.pi / sym
        self.colfunc = colfunc
        self.bgcol = hsl_color(0, 0, -0.5)
        self.horizon = 320
        self.preferred_frames = 640
        self.proj2D3D = "halfsphere"

    def create(self):
        col = self.colfunc(self.time)
        speed = m.exp((random() - 0.5) * 1.1) / self.preferred_fps * 0.08
        rad = 0.1 + random() * 0.3
        angle = random() * 2 * m.pi
        offset = random() - 0.5
        vec = (m.sin(angle), m.cos(angle))
        cent = (-vec[0] * 1.5 + vec[1] * offset, -vec[1] * 1.5 - vec[0] * offset + 0.5)
        vel = (vec[0] * speed, vec[1] * speed)
        nstep = int(3.0 / speed)
        shape = Blob(cent, rad, col)
        #shape = Ellipse(cent, 0.0, rad, rad, col)
        shape.set_depth(0.0)
        shape.set_speed(vel[0], vel[1])
        shape.duetime = self.time + nstep
        return shape

    def get_color(self, coord, ind):
        r = m.sqrt(coord[0] ** 2 + coord[1] ** 2)
        a = m.atan2(coord[1], coord[0])
        a = abs((a % (2 * self.symang)) - self.symang)
        return super(SoftKaleidoScene, self).get_color((m.sin(a) * r, m.cos(a) * r), ind)

    def blend_colors(self, colors):
        cols = colors
        br = list(map(lambda rgb: color_brightness(*rgb), colors))
        maxbr = max(br) if br else 0.0
        sumbr = sum(br) if br else 0.0
        bgbr = color_brightness(*self.bgcol)
        if maxbr < bgbr:
            delta = (bgbr - maxbr) / bgbr
            sumbr += (bgbr - maxbr)
            maxbr = bgbr
            cols = cols + [tuple(map(lambda x: x*delta, self.bgcol))]
        sumcol = tuple(map(lambda *args: max(0, min(255, int(round(sum(args) * maxbr / sumbr)))), *cols))
        return sumcol

ctr = setup_control()
ctr.adjust_layout_aspect(1.0)  # How many times wider than high is the led installation?
SoftKaleidoScene(ctr, 5, random_hsl_color_func(light=0.0)).launch_movie()



"""
Day 16 - Wheel of whites

In my oppinion, the "Sunset" effect on the app is quite ugly and distasteful
with its saturated colors. It is not easy to save it. Nevertheless, to show
that this type of rotating pattern can be generated with xled_plus too, I
created this effect of tinted whites, or very light watercolors that could
well each have passed for white if they were not next to each other, slowly
rotating around a neutral white center.
"""

from xled_plus.samples.sample_setup import *

class WheelEffect(Effect):
    def __init__(self, ctr, cols, folds, speed):
        super(WheelEffect, self).__init__(ctr)
        self.cols = cols
        self.folds = folds
        self.ncols = len(cols) * folds
        self.speedfactor = float(speed) / self.preferred_fps
        self.preferred_frames = int(len(cols) * self.preferred_fps / speed)
        self.white = hsl_color(0.0, 0.0, 1.0)
    def getcolor(self, pos):
        a = m.atan2(pos[1], pos[0])
        r = m.sqrt(pos[0]**2 + pos[1]**2)
        n = round(a * self.ncols / (2 * m.pi) + self.tm * self.speedfactor)
        col = self.cols[int(n) % len(self.cols)]
        if r < 0.3:
            return self.white
        else:
            return col
    def reset(self, numframes):
        self.tm = 0
    def getnext(self):
        pat = self.ctr.make_layout_pattern(self.getcolor, "centered")
        self.tm += 1
        return pat

ctr = setup_control()
cols = [hsl_color(0.3995, 1.0000, 0.6110), hsl_color(0.6609, 1.0000, 0.7104), hsl_color(0.1321, 1.0000, 0.6374), hsl_color(0.4466, 1.0000, 0.5911), hsl_color(0.0395, 1.0000, 0.5389)]
WheelEffect(ctr, cols, 1, 0.3).launch_movie()



from .sample_setup import *


ctr = setup_control()
ctr.show_movie(
    ctr.make_func_movie(
        120,
        lambda t: ctr.make_layout_pattern(
            lambda pos: hsl_color((pos[1] - t / 120.0) % 1.0, 1.0, 0.0)
        ),
    ),
    8,
)



"""
xled_plus.xled_colorpicker
~~~~~~~~~~~~~~~~~~~~~~~~~~

This is an example application of the color picker for showing and
setting colors on the leds.

Create an XledColorPicker with a HighControlInterface as argument.
A window with a colored sphere will appear. Hover over the sphere
to watch colors. Click on a color to upload it to the leds as a movie.
Drag the surface of the sphere to rotate it.

When started from the shell, eg with 'python -m xled_plus.xled_colorpicker'
it will look on the netowork for any connected leds. If you know the
ip-address of your leds, you can provide it as an extra argument and it
will not have to search for it.
"""

from colorsphere.colorsphere import ColorPicker
from xled_plus.ledcolor import hsl_color, get_color_style, set_color_style
from xled_plus.highcontrol import HighControlInterface
from xled.discover import discover
import sys


class XledColorPicker:
    rtmode = False
    outermode = False
    printrgb = False
    printhsl = False
    noset = False

    def __init__(self, ctr):
        self.ctr = ctr

    def on_click(self, hsl, event):
        if hsl:
            if not self.noset:
                pat = self.ctr.make_solid_pattern(hsl_color(*hsl))
                id = self.ctr.upload_movie(self.ctr.to_movie(pat), 1, force=True)
                self.ctr.set_movies_current(id)
                self.outermode = 'movie'
            if self.printrgb and self.printhsl:
                args = hsl + (hsl_color(*hsl), )
                print("HSL: ({:.4f}, {:.4f}, {:.4f})  RGB: {}".format(*args))
            elif self.printrgb:
                print(hsl_color(*hsl))
            elif self.printhsl:
                print("({:.4f}, {:.4f}, {:.4f})".format(*hsl))

    def on_move(self, hsl, event):
        if hsl:
            if not self.rtmode:
                self.outermode = self.ctr.get_mode()['mode']
            pat = self.ctr.make_solid_pattern(hsl_color(*hsl))
            self.ctr.show_rt_frame(self.ctr.to_movie(pat))
            self.rtmode = True
        else:
            if self.rtmode:
                if self.outermode:
                    self.ctr.set_mode(self.outermode)
                self.rtmode = False

    def colorstyle_changed(self, style):
        if type(style) in [list, tuple]:
            for s in style:
                set_color_style(s)
        else:
            set_color_style(style)

    def exit_event_loop(self, *args):
        self.cp.win.fig.canvas.stop_event_loop()

    def set_color_style(self, style):
        self.cp.sphere.set_color_style(style)

    def launch(self, from_shell=False, printrgb=False, printhsl=False, noset=False):
        self.printrgb = printrgb
        self.printhsl = printhsl
        self.noset = noset
        self.cp = ColorPicker(self.on_click, self.on_move, name="Xled Color Picker")
        self.cp.sphere.set_color_style(get_color_style())
        self.cp.sphere.color_style_callbacks.append(self.colorstyle_changed)
        if from_shell:
            self.cp.win.add_close_callback(self.exit_event_loop)
            self.cp.win.fig.canvas.start_event_loop(0)


if __name__ == '__main__':

    noset = False
    rgb = False
    hsl = False
    host = False
    for arg in sys.argv:
        if arg == "noset":
            noset = True
        elif arg == "rgb":
            rgb = True
        elif arg == "hsl":
            hsl = True
        elif arg.replace(".","").isdigit():
            host = arg
    if not host:
        host = discover().ip_address
    ctr = HighControlInterface(host)
    XledColorPicker(ctr).launch(from_shell=True, printhsl=hsl, printrgb=rgb, noset=noset)



"""
Day 11 - Gradients of yellow

Gradients of yellow, green and orange sweeping diagonally over the lights.
Uses the layout, and for best effect you may have to adjust the aspect ratio:
The parameter to adjust_layout_aspect should be set to width/height of the lights.
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
ctr.adjust_layout_aspect(1.0)  # How many times wider than high is the led installation?
cols = [hsl_color(0.375, 0.2, 0.0), hsl_color(0.375, 1.0, 0.0), hsl_color(0.28, 1.0, 0.0), hsl_color(0.375, 0.2, 0.0), hsl_color(0.375, 1.0, 0.0), hsl_color(0.5, 1.0, 0.0)]
GradientSequence(ctr, cols, speed=0.04, folds=0.5, angle=45).launch_movie()



"""
Day 28 - Flowers

Different colored flowers that bloom from the center. 
Similar principle as the Kaleidoscope, but grow pieces from the center.
Again, looks best on a 2D layout, and with correctly adjusted aspect:
Change the number in the call to adjust_layout_aspect in the code below.
"""

from xled_plus.samples.sample_setup import *

class FlowerScene(MovingShapesScene):
    def __init__(self, ctr, sym):
        super(FlowerScene, self).__init__(ctr)
        self.freq = 0.6
        self.symang = m.pi / sym
        self.bgcol = hsl_color(0.25, 1.0, -0.5)
        self.lasthue = 0.25
        self.horizon = 360
        self.preferred_frames = 720
        self.proj2D3D = "halfsphere"
        self.ind = 0

    def makecolorfunc(self, sh):
        hue = (random()*0.74 + 0.36) % 1.0
        while abs(hue - self.lasthue) < 0.1:
            hue = (random()*0.74 + 0.36) % 1.0
        self.lasthue = hue
        sat = random()*0.5 + 0.5
        aa = 0.2
        ss = 0.05 / self.preferred_fps
        def func(pos, ind):
            tm = self.time - sh.crtime
            if tm * ss < aa:
                return hsl_color(hue, sat, 0.0)
            else:
                delta = 2.0 - 2.0 * aa / (tm * ss)
                rad = (pos[0]**2 + pos[1]**2) ** 0.5
                satres = rad * delta / (tm * ss) + 1.0 - delta
                return hsl_color(hue, max(0.0, min(1.0, sat * satres)), 0.0)
        return func

    def reshapefunc(self, sh):
        fact = (random() - 0.4) * 0.7
        aa = 0.2
        ss = 0.05 / self.preferred_fps
        def func():
            tm = self.time - sh.crtime
            if tm * ss < aa:
                sh.cent = (0.0, 0.0)
                sh.rad1 = tm * ss
                sh.rad2 = sh.rad1
            else:
                st = tm - aa / ss
                sh.cent = (sh.speed[0] * ss * st * 0.5, sh.speed[1] * ss * st * 0.5)
                sh.rad2 = aa + st * ss * 0.5
                sh.rad1 = aa + st * fact * ss * 0.5
        return func

    def create(self):
        angle = (random() * 3.0 - 0.5) * self.symang
        vec = (m.sin(angle), m.cos(angle))
        shape = Ellipse((0.0, 0.0), angle, 0.0, 0.0, False)
        shape.crtime = self.time
        shape.color = self.makecolorfunc(shape)
        shape.reshape = self.reshapefunc(shape)
        shape.set_depth(-self.ind)
        shape.set_speed(vec[0], vec[1])
        self.ind += 1
        return shape

    def update(self, step):
        if self.time % 30 == 0:
            if self.record:
                sh = self.create()
                self.add_shape(sh)
                self.pre_shapes.append((self.time + self.preferred_frames, sh))
            elif self.pre_shapes and self.pre_shapes[0][0] <= self.time:
                sh = self.pre_shapes[0][1]
                sh.crtime = self.time
                sh.set_depth(-self.ind)
                self.ind += 1
                self.add_shape(sh)
                self.pre_shapes = self.pre_shapes[1:]
            else:
                self.add_shape(self.create())
        for sh in self.shapes:
            sh.reshape()
        if len(self.shapes) > 12:
            self.shapes = self.shapes[:-1]
        self.time += 1
        # super(FlowerScene, self).update(step)

    def get_color(self, coord, ind):
        r = m.sqrt(coord[0] ** 2 + coord[1] ** 2)
        a = m.atan2(coord[1], coord[0])
        a = abs((a % (2 * self.symang)) - self.symang)
        return super(FlowerScene, self).get_color((m.sin(a) * r, m.cos(a) * r), ind)

    def blend_colors(self, colors):
        cols = colors
        br = list(map(lambda rgb: color_brightness(*rgb), colors))
        maxbr = max(br) if br else 0.0
        sumbr = sum(br) if br else 0.0
        bgbr = color_brightness(*self.bgcol)
        if maxbr < bgbr:
            delta = (bgbr - maxbr) / bgbr
            sumbr += (bgbr - maxbr)
            maxbr = bgbr
            cols = cols + [tuple(map(lambda x: x*delta, self.bgcol))]
        sumcol = tuple(map(lambda *args: max(0, min(255, int(round(sum(args) * maxbr / sumbr)))), *cols))
        return sumcol

if __name__ == '__main__':
    ctr = setup_control()
    ctr.adjust_layout_aspect(1.0)  # How many times wider than high is the led installation?
    eff = FlowerScene(ctr, 6)
    oldmode = ctr.get_mode()["mode"]
    eff.launch_rt()
    print("Started continuous effect - press Return to stop it")
    input()
    eff.stop_rt()
    ctr.set_mode(oldmode)




"""
Day 6 - Flash with mixed colors

Light colors starting with a flash and then fading out.
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
SparkleEffect(ctr, 4, random_color_func(light=0.2), flashlight_func(16, 16)).launch_movie()



"""
Day 24 - Christmas Tree Balls

Colored balls moving against a glittering white background.
Merry Christmas!
"""

from xled_plus.samples.sample_setup import *

class ChristmasBallsScene(MovingShapesScene):
    def __init__(self, ctr, cols):
        super(ChristmasBallsScene, self).__init__(ctr)
        self.freq = 0.5
        self.colors = cols
        self.whites = [hsl_color(0.4, 0.5, 0.2), hsl_color(0.1, 0.5, 0.2), hsl_color(0.0, 0.0, 0.2)]
        self.blink = hsl_color(0.0, 0.0, 1.0)
        self.bgpat = make_random_select_color_pattern(ctr, self.whites)
        self.preferred_frames = 640
        self.proj2D3D = "cylshell"

    def create(self):
        tp = int(random() * len(self.colors))
        speed = 0.02
        angle = random() * 2 * m.pi
        offset = (random() - 0.5) * 1.6
        vec = (m.sin(angle), m.cos(angle))
        cent = (-vec[0] * 1.5 + vec[1] * offset, -vec[1] * 1.5 - vec[0] * offset)
        vel = (vec[0] * speed, vec[1] * speed)
        nstep = int(3.0 / speed)
        sh = Ellipse(cent, 0.0, 0.16, 0.16, self.colors[tp])
        sh.set_speed(vel[0], vel[1])
        sh.set_depth(0.0)
        sh.duetime = self.time + nstep
        return sh
        
    def getnext(self):
        pat1 = super(ChristmasBallsScene, self).getnext()
        patbg = sprinkle_pattern(self.ctr, self.bgpat, [self.blink], 0.05*self.ctr.num_leds)
        vec = self.getoccupancy()
        pat = [pat1[i] if vec[i] else patbg[i] for i in range(self.ctr.num_leds)]
        return pat

ctr = setup_control()
ctr.adjust_layout_aspect(1.0)  # How many times wider than high is the led installation?
cols = [hsl_color(0.625, 1.0, -0.15), hsl_color(0.26, 1.0, -0.11), hsl_color(0.42, 1.0, -0.05), hsl_color(0.54, 1.0, -0.15)]
# cols = [hsl_color(0.44, 1.0, -0.21), hsl_color(0.44, 1.0, 0.11), hsl_color(0.625, 1.0, -0.17), hsl_color(0.625, 1.0, 0.11)]
ChristmasBallsScene(ctr, cols).launch_movie()



"""
Day 8 - Flash with pair of slowly changing colors

Sparkling effect, starting with a flash and fading out, just like Day6,
but now with two simultaneous colors, slowly traversing the spectrum. 
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
eff = SparkleEffect(ctr, 4, circular_color_func(240, [0.0, 0.25]), flashlight_func(16, 16))
eff.preferred_frames = 240
eff.launch_movie()



"""
Day 4 - White tinted gradient

Slowly changing gradient between two complementary light colors,
such that the middle stays white and the colors above and below
traverses the spectrum on opposite sides.

This effect uses the physical layout of the leds. It only uses
one coordinate though (y-axis), so it should work for 2D, 3D, or
linear layouts. (Older firmware do not store the layout on the leds.
For them the effect will follow the string instead.)
"""

from xled_plus.samples.sample_setup import *

def tintedwhite(t, pos):
    y = pos[1] if len(pos) > 1 else pos[0]
    return hsl_color(t%1.0 if y<0.5 else (t + 0.5)%1.0, 1.0, 1.0 - abs(y - 0.5)*1.6)

ctr = setup_control()
mov = ctr.make_func_movie(120, lambda t: ctr.make_layout_pattern(lambda pos: tintedwhite(t / 120.0, pos)))
ctr.show_movie(mov, 0.5)



"""
xled_plus.xled_picture
~~~~~~~~~~~~~~~~~~~~~~

Author: Anders Holst (anders.holst@ri.se), 2021

An effect that shows a picture (possibly animated) on your leds.
The PictureEffect object takes as parameters: the HighControlInterface
for the leds, the full filename of the picture, and optionally a
parameter `fit` which can be: 'stretch' (for stretching the axes to fit
the layout bounds), 'small' (to fit the entire picture inside the layout,
leaving black edges), 'large' (to fill up the entire layout, truncating
parts that does not fit), and 'medium' (which is a geometrical average
between small and large). 'stretch' is the default.
If the picture is an animated gif, it will become a move, otherwise it is
a static effect.

It can be invoked from the shell as:
`python -m xled_plus.xled_picture [host] filename`
where [host] is an optional ip number of the leds.

This module uses the Python Image Processing Library (pillow), to read
the picture from file.
"""

from PIL import Image
from xled_plus.effect_base import Effect
from xled_plus.ledcolor import image_to_led_rgb
from xled_plus.highcontrol import HighControlInterface
from xled.discover import discover
import sys


class PictureEffect(Effect):
    def __init__(self, ctr, fname, fit='stretch'):
        # fit can be: 'stretch', 'small', 'large', 'medium'
        super(PictureEffect, self).__init__(ctr)
        self.im = Image.open(fname)
        self.xmid = (self.im.size[0]-1) / 2.0
        self.ymid = (self.im.size[1]-1) / 2.0
        if fit == 'stretch':
            self.xscale = self.im.size[0]-1
            self.yscale = self.im.size[1]-1
        else:
            bounds = ctr.get_layout_bounds()
            xdiff = bounds["bounds"][0][1] - bounds["bounds"][0][0]
            if xdiff == 0.0:
                xdiff = 1.0
            if fit == 'small':
                fact = max((self.im.size[0]-1) / xdiff, self.im.size[1]-1)
            elif fit == 'large':
                fact = min((self.im.size[0]-1) / xdiff, self.im.size[1]-1)
            else:
                fact = ((self.im.size[0]-1) * (self.im.size[0]-1) / xdiff) ** 0.5
            self.xscale = xdiff * fact
            self.yscale = fact
        if "is_animated" in dir(self.im) and self.im.is_animated:
            self.preferred_fps = 100.0 / self.im.info["duration"]
            self.preferred_frames = self.im.n_frames
        else:
            self.preferred_fps = 1
            self.preferred_frames = 1

    def get_color(self, pos):
        coord = (int(round((pos[0] - 0.5) * self.xscale + self.xmid)),
                 int(round((0.5 - pos[1]) * self.yscale + self.ymid)))
        if coord[0] >= 0 and coord[0] < self.im.size[0] and coord[1] >= 0 and coord[1] < self.im.size[1]:
            if self.im.mode == 'P':
                pix = self.im.getpixel(coord)
                rgb = self.im.getpalette()[pix*3:pix*3+3]
            elif self.im.mode == 'L':
                rgb = [self.im.getpixel(coord)] * 3
            else:
                rgb = self.im.getpixel(coord)[0:3]
        else:
            rgb = (0, 0, 0)
        return image_to_led_rgb(*rgb)

    def reset(self, numframes):
        self.index = -1

    def getnext(self):
        if "is_animated" in dir(self.im) and self.im.is_animated:
            self.index += 1
            self.im.seek(self.index % self.im.n_frames)
        return self.ctr.make_layout_pattern(self.get_color, style="square")


if __name__ == '__main__' and len(sys.argv) > 1:

    if sys.argv[1].replace(".","").isdigit() and len(sys.argv) > 2:
        host = sys.argv[1]
        file = sys.argv[2]
    else:
        host = discover().ip_address
        file = sys.argv[1]
    ctr = HighControlInterface(host)
    PictureEffect(ctr, file).launch_movie()



"""
Day 18 - Moving shapes

Colored geometric shapes moving over the leds.
Looks best on a 2D layout, and with correctly adjusted aspect: Change
the number in the call to adjust_layout_aspect in the code below.

I promised to not make just flashy but instead tasteful patterns.
This time maybe I failed, by trying to show a new kind of effect
just to prove it can be done, rather than for the estecics of it...
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
ctr.adjust_layout_aspect(1.0)  # How many times wider than high is the led installation?
eff = MovingShapesScene(ctr)
eff.preferred_frames = 720
eff.launch_movie()



"""
Day 9 - Pulselights green

Smoothly pulsing lights with random colors from the green spectrum.
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
colfunc = random_color_func(hue=[0.16, 0.38], light=0.0)
nsparks = 0.5 * ctr.num_leds / 60.0
eff = SparkleEffect(ctr, nsparks, colfunc, pulselight_func(24, 12, 24))
eff.preferred_fps = 12
eff.preferred_frames = 180
eff.launch_movie()



"""
Day 26 - Random walk with complementary colors

Variant of the Day 25 effect, with a gradient between a random walk
color and its complement. 
This is again a continuous effect, not a movie, so when you stop the script
the effect stops. 
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
oldmode = ctr.get_mode()["mode"]
eff = MeanderingTandemSequence(ctr, gap=0.5, fixangle=0.0)
eff.launch_rt()
print("Started continuous effect - press Return to stop it")
input()
eff.stop_rt()
ctr.set_mode(oldmode)



# -*- coding: utf-8 -*-

"""
xled_plus.multicontrol
~~~~~~~~~~~~~~~~~~~~~~

Handle multiple connected lights as one unit.
Create a MultiHighControlInterface by giving it a list of ip-addresses to
the devices that are connected, with the master device first. Then, use it
in the same way as a normal HighControlInterface.

"""

from __future__ import absolute_import

import io
import struct
import binascii
import time
import uuid

import struct
import base64

from xled.control import ControlInterface
from xled_plus.highcontrol import HighControlInterface


def pick_master_and_slaves(hostlst):
    ctrlst = [ControlInterface(ip) for ip in hostlst]
    master = False
    slaves = []
    for ctr in ctrlst:
        sync = ctr.get_led_movie_config()['sync']
        if sync['mode'] == 'master':
            assert not master, "Only one master device allowed in the group"
            master = ctr
        elif sync['mode'] == 'slave':
            slaves.append(ctr)
        else:
            assert len(hostlst) == 1, "One device does not belong to the group"
            master = ctr
    assert master, "No master device in the group"
    return master, slaves


class MultiHighControlInterface(HighControlInterface):
    """
    High level interface to control a group of joined devices in sync
    """

    def __init__(self, hostlst):
        master, slaves = pick_master_and_slaves(hostlst)
        super(MultiHighControlInterface, self).__init__(master.host)
        self.ctrlst = slaves
        info = self.get_device_info()
        self.family = info["fw_family"] if "fw_family" in info else "D"
        self.led_bytes = info["bytes_per_led"] if "bytes_per_led" in info else 3
        self.led_profile = info["led_profile"] if "led_profile" in info else "RGB"
        self.version = tuple(map(int, self.firmware_version()["version"].split(".")))
        self.nledslst = [ctr.get_device_info()["number_of_led"] for ctr in self.ctrlst]
        for ctr in self.ctrlst:
            ctr._udpclient = self.udpclient
        self.num_leds = sum(self.nledslst)
        self.string_config = [{'first_led_id': 0, 'length': self.num_leds}]
        self.layout = False
        self.layout_bounds = False
        self.last_mode = None
        self.last_rt_time = 0
        self.curr_mode = self.get_mode()["mode"]

    def split_movie(self, movie):
        # return a list of one movie per connected device
        lst = [io.BytesIO() for ctr in self.ctrlst]
        blens = [nleds * self.led_bytes for nleds in self.nledslst]
        totlen = self.num_leds * self.led_bytes
        num = movie.seek(0, 2) // totlen
        movie.seek(0)
        for i in range(num):
            for mov, blen in zip(lst, blens):
                mov.write(movie.read(blen))
        for mov in lst:
            mov.seek(0)
        return lst

    def show_movie(self, movie_or_id, fps=None):
        """
        Either starts playing an already uploaded movie with the provided id,
        or uploads a new movie and starts playing it at the provided frames-per-second.
        Note: if the movie do not fit in the remaining capacity, the old movie list is cleared.
        Switches to movie mode if necessary.
        The movie is an object suitable created with to_movie or make_func_movie.

        :param movie_or_id: either an integer id or a file-like object that points to movie
        :param fps: frames per second, or None if a movie id is given
        """
        if isinstance(movie_or_id, int) and fps is None:
            if self.family == "D" or self.version < (2, 5, 6):
                if movie_or_id != 0:
                    return False
            else:
                movies = self.get_movies()["movies"]
                if movie_or_id in [entry["id"] for entry in movies]:
                    self.set_movies_current(movie_or_id)
                else:
                    return False
            if self.curr_mode != "movie":
                self.set_mode("movie")
        else:
            assert fps
            self.set_mode("off")
            self.upload_movie(movie_or_id, fps, force=True)
            self.set_mode("movie")
        return True

    def upload_movie(self, movie, fps, force=False):
        """
        Uploads a new movie with the provided frames-per-second.
        Note: if the movie does not fit in the remaining capacity, and force is
        not set to True, the function just returns False, in which case the user
        can try clear_movies first.
        Does not switch to movie mode, use show_movie instead for that.
        The movie is an object suitable created with to_movie or make_func_movie.
        Returns the new movie id, which can be used in calls to show_movie or
        show_playlist.

        :param movie: a file-like object that points to movie
        :param fps: frames per second, or None if a movie id is given
        :param bool force: if remaining capacity is too low, previous movies will be removed
        :rtype: int
        """
        numframes = movie.seek(0, 2) // (self.led_bytes * self.num_leds)
        movielst = self.split_movie(movie)
        if self.family == "D" or self.version < (2, 5, 6):
            for ctr, mov, nled in zip(self.ctrlst, movielst, self.nledslst):
                ctr.set_led_movie_config(1000 // fps, numframes, nled)
                ctr.set_led_movie_full(mov)
            return 0
        else:
            res = self.get_movies()
            capacity = res["available_frames"] - 1
            if numframes > capacity or len(res["movies"]) > 15:
                if force:
                    if self.curr_mode == "movie" or self.curr_mode == "playlist":
                        self.set_mode("effect")
                    self.delete_movies()
                else:
                    return False
            if self.curr_mode == "movie":
                oldid = self.get_movies_current()["id"]
            for ctr, mov, nled in zip(self.ctrlst, movielst, self.nledslst):
                res = ctr.set_movies_new(
                    "",
                    str(uuid.uuid4()),
                    self.led_profile.lower() + "_raw",
                    nled,
                    numframes,
                    fps,
                )
                ctr.set_movies_full(mov)
            if self.curr_mode == "movie":
                self.set_movies_current(oldid)  # Dont change currently shown movie
            return res["id"]

    def show_rt_frame(self, frame):
        """
        Uploads a frame as the next real time frame, and shows it.
        Switches to rt mode if necessary.
        The frame is either a pattern or a one-frame movie

        :param frame: a pattern or file-like object representing the frame
        """
        if self.is_pattern(frame):
            frame = self.to_movie(frame)
        framelst = self.split_movie(frame)
        if self.curr_mode != "rt" or self.last_rt_time + 50.0 < time.time():
            for ctr in self.ctrlst:
                ctr.set_mode("rt")
        else:
            self.last_rt_time = time.time()
        for ctr, mov, nled in zip(self.ctrlst, framelst, self.nledslst):
            self.udpclient.destination_host = ctr.host
            if self.family == "D":
                ctr.set_rt_frame_socket(mov, 1, nled)
            elif self.version < (2, 4, 14):
                ctr.set_rt_frame_socket(mov, 2)
            else:
                ctr.set_rt_frame_socket(mov, 3)
        self.udpclient.destination_host = self.host

    def clear_movies(self):
        """
        Removes all uploaded movies and any playlist.
        If the current mode is 'movie' or 'playlist' it switches mode to 'effect'
        """
        if self.curr_mode == "movie" or self.curr_mode == "playlist":
            self.set_mode("effect")
        if self.family == "D" or self.version < (2, 5, 6):
            # No list of movies to remove in this version,
            # but disable movie mode until new movie is uploaded
            for i, ctr in enumerate(self.ctrlst):
                ctr.set_led_movie_config(1000, 0, self.nledslst[i])
        else:
            # The playlist is removed automatically when movies are removed
            for ctr in self.ctrlst:
                ctr.delete_movies()

    def get_led_layout(self):
        res = super(MultiHighControlInterface, self).get_led_layout()
        for ctr in self.ctrlst[1:]:
            tmp = ctr.get_led_layout()
            res["coordinates"].extend(tmp["coordinates"])
        return res





"""
Day 31 - Fireworks (version A)

My take on the fireworks theme. Colored blobs growing and sparkling.
This version gives one explosion at a time at controlled intervalls.
Suits best for smaller nonregular installations, eg a christmas tree.
Also check the version in day31b.
"""

from xled_plus.samples.sample_setup import *


class Fireball(Shape):
    def __init__(self, cent, minrad, maxrad, speed, huesat):
        super(Fireball, self).__init__(cent, 0.0)
        self.rad = minrad
        self.minrad = minrad
        self.maxrad = maxrad
        self.speed = speed
        self.huesat = huesat

    def is_inside(self, coord):
        return (
            sum(map(lambda x1, x2: (x1 - x2) ** 2, self.cent, coord)) <= self.rad ** 2
        )

    def get_color(self, coord, ind):
        dist2 = (
            sum(map(lambda x1, x2: (x1 - x2) ** 2, self.cent, coord)) / self.rad ** 2
        )
        if dist2 > 1.0:
            return False
        elif self.rad == self.minrad:
            return hsl_color(0.0, 0.0, 1.0)
        else:
            (hue, sat) = self.huesat
            p = (self.rad - self.minrad) / (self.maxrad - self.minrad)
            rnd = random()
            if rnd < 0.1 * (1.2 - p):
                light = 1.0
            elif rnd < 0.2 * (1.2 - p):
                light = 0.0
            else:
                light = 2.0 * (1.0 - p) ** 2 - 1.0
            return hsl_color(hue, sat, light)

    def update(self, step):
        self.rad += self.speed * step


class FireworksScene(MovingShapesScene):
    def __init__(self, ctr):
        super(FireworksScene, self).__init__(ctr)
        self.freq = 0.0
        self.horizon = 100
        self.preferred_frames = 700
        self.preferred_fps = 20
        self.proj2D3D = "halfsphere"

    def create(self):
        cent = (0.0, 0.0)
        shape = Fireball(cent, 0.10, 1.0, 0.01, (random(), 1.0 - random()**2))
        shape.duetime = self.time + 0.9/0.01
        return shape

    def update(self, step):
        for sh in reversed(self.shapes):
            if sh.duetime < self.time:
                self.shapes.remove(sh)
        if self.time >= self.crtime:
            if self.record:
                sh = self.create()
                self.add_shape(copy(sh))
                sh.duetime += self.preferred_frames
                self.crtime = self.time + 100
                self.pre_shapes.append((self.crtime + self.preferred_frames, sh))
            elif self.replay:
                if self.pre_shapes:
                    self.add_shape(self.pre_shapes[0][1])
                    self.crtime = self.pre_shapes[0][0]
                    self.pre_shapes = self.pre_shapes[1:]
            else:
                self.add_shape(self.create())
                self.crtime = self.time + 100
                if self.pre_shapes and self.crtime >= self.pre_shapes[0][0]:
                    self.replay = True
        for sh in self.shapes:
            sh.update(step)
        self.time += step

if __name__ == '__main__':
    ctr = setup_control()
    ctr.adjust_layout_aspect(1.0)  # How many times wider than high is the led installation?
    eff = FireworksScene(ctr)
    oldmode = ctr.get_mode()["mode"]
    eff.launch_rt()
    print("Started continuous effect - press Return to stop it")
    input()
    eff.stop_rt()
    ctr.set_mode(oldmode)



"""
Day 1 - Twinkling stars

A simple effect to start with. It is similar to the Star effect in the
Twinkly app. However, in the app all stars have the same purple-tinted
sharp white color ("Twinkly-white"). Here instead the stars have
different color temperatures, from warm, through neutral white to cold
- just like real stars.
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
SparkleStars(ctr).launch_movie()



"""
Day 31 - Fireworks (version B)

My take on the fireworks theme. Colored blobs growing and sparkling.
This version gives explosions at random times and positions.
Suits best for rather large and dense 2D installations.
Also check the version in day31a.
"""

from xled_plus.samples.sample_setup import *


class Fireball(Shape):
    def __init__(self, cent, minrad, maxrad, speed, huesat):
        super(Fireball, self).__init__(cent, 0.0)
        self.rad = minrad
        self.minrad = minrad
        self.maxrad = maxrad
        self.speed = speed
        self.huesat = huesat

    def is_inside(self, coord):
        return (
            sum(map(lambda x1, x2: (x1 - x2) ** 2, self.cent, coord)) <= self.rad ** 2
        )

    def get_color(self, coord, ind):
        dist2 = (
            sum(map(lambda x1, x2: (x1 - x2) ** 2, self.cent, coord)) / self.rad ** 2
        )
        if dist2 > 1.0:
            return False
        elif self.rad == self.minrad:
            return hsl_color(0.0, 0.0, 1.0)
        else:
            (hue, sat) = self.huesat
            p = (self.rad - self.minrad) / (self.maxrad - self.minrad)
            rnd = random()
            if rnd < 0.1 * (1.2 - p):
                light = 1.0
            elif rnd < 0.2 * (1.2 - p):
                light = 0.0
            else:
                light = 2.0 * (1.0 - p) ** 2 - 1.0
            return hsl_color(hue, sat, light)

    def update(self, step):
        self.rad += self.speed * step


class FireworksScene(MovingShapesScene):
    def __init__(self, ctr):
        super(FireworksScene, self).__init__(ctr)
        self.freq = 0.8
        self.horizon = 60
        self.preferred_frames = 720
        self.preferred_fps = 20
        self.proj2D3D = "cylshell"

    def create(self):
        cent = (random()*1.6 - 0.8, random()*1.6 - 0.8)
        shape = Fireball(cent, 0.15, 0.63, 0.008, (random(), 1.0))
        shape.duetime = self.time + 0.48/0.008
        return shape


if __name__ == '__main__':
    ctr = setup_control()
    ctr.adjust_layout_aspect(1.0)  # How many times wider than high is the led installation?
    eff = FireworksScene(ctr)
    oldmode = ctr.get_mode()["mode"]
    eff.launch_rt()
    print("Started continuous effect - press Return to stop it")
    input()
    eff.stop_rt()
    ctr.set_mode(oldmode)



from xled.discover import discover
from xled_plus.highcontrol import HighControlInterface
from xled_plus.ledcolor import *
from xled_plus.shapes import *
from sys import argv
import re

def isipaddress(txt):
    return re.match("([0-9]+)\.([0-9]+)\.([0-9]+)\.([0-9]+)", txt)

def makecolorlist(n):
    gs = (1.25**0.5 - 0.5)
    return [hsl_color((0.5 + gs * i) % 1.0, 1.0, 0.0) for i in range(n)]

if len(argv) == 3 and isipaddress(argv[1]):
    host = argv[1]
    txt = argv[2]
elif len(argv) == 3 and isipaddress(argv[2]):
    host = argv[2]
    txt = argv[1]
elif len(argv) == 2:
    txt = argv[1]
    host = discover().ip_address
else:
    print('Usage: python -m xled_plus.show_text [ip-address] "text"')
    quit()

ctr = HighControlInterface(host)
ctr.adjust_layout_aspect(1.0)
eff = RunningText(ctr, txt.upper(), makecolorlist(len(txt)), size=0.6, speed=0.5)
eff.launch_movie()



"""
Day 15 - Looping lights, blue, purple, orange

Same principle as yesterday with different colors.
Because I think it looks nice.
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
nsparks = 0.4 * ctr.num_leds / 32.0
colfunc = selected_color_func([(0.9, 1.0), (0.1, 1.0), (0.5, 1.0)])
SparkleEffect(ctr, nsparks, colfunc, looplight_func(16, 16)).launch_movie()



"""
Day 25 - Random walk in color space

Solid color that slowly changes randomly.
This is a continuous effect, not a movie, so when you stop the script
the effect stops. 
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
oldmode = ctr.get_mode()["mode"]
eff = ColorMeanderEffect(ctr, "solid")
eff.launch_rt()
print("Started continuous effect - press Return to stop it")
input()
eff.stop_rt()
ctr.set_mode(oldmode)



"""
Day 23 - Crystal

Glittering white crystal, yet shimmering in all colors at once.
"""

from xled_plus.samples.sample_setup import *
from random import shuffle

class Crystal(RotateEffect):
    def __init__(self, ctr):
        saturations = [0.4, 0.0, 0.0, 0.2, 0.0, 0.0]
        folds = len(saturations)
        nleds = ctr.num_leds
        pat = ctr.make_func_pattern(
            lambda i: hsl_color((i*folds / float(nleds))%1.0, saturations[i*folds // nleds], 0.0),
            circular=True,
        )
        perm = list(range(nleds))
        shuffle(perm)
        self.nsparks = 0.02 * nleds
        self.white = hsl_color(0.0, 0.0, 1.0)
        super(Crystal, self).__init__(ctr, pat, perm, step=2)
        self.preferred_fps = 12

    def getnext(self):
        pat1 = super(Crystal, self).getnext()
        return sprinkle_pattern(self.ctr, pat1, [self.white], self.nsparks)

ctr = setup_control()
Crystal(ctr).launch_movie()



"""
xled_plus.pattern
~~~~~~~~~~~~

Author: Anders Holst (anders.holst@ri.se), 2021

This module contains a number of utility functions useful when creating
patterns, as well as several functions for creating various patterns.
The latter are meant primarily as a collection of examples. It should be
easy to construct many more types of patterns with just a few lines of code.

All the functions for creating patterns take a parameter ctr, which is a
HighControlInterface connected to the led lights on which the pattern will fit.
"""

from xled_plus.ledcolor import hsl_color, rgb_color
import random
import math as m


# Some utility functions


def randomdiscrete(probs):
    """
    Takes a list of probabilities that should sum to one, and returns a random
    list index according to the probabilities in the list.
    """
    n = len(probs) - 1
    acc = 0.0
    ind = -1
    r = random.random()
    while acc <= r and ind < n:
        ind += 1
        acc += probs[ind]
    return ind


def randompoisson(lam):
    """
    Returns a random number from a Poisson distribution with parameter lam.
    """
    lp = 0.0
    k = -1
    while lp < lam:
        lp -= m.log(random.random())
        k += 1
    return k


def dimcolor(rgb, prop):
    """
    Dims a given rgb color to prop, which should be in the interval 0.0 - 1.0.
    """
    return tuple(map(lambda x: int(round(x * prop)), rgb))


def blendcolors(rgb1, rgb2, prop):
    """
    Blends two rgb colors so that prop comes from rgb2 and (1-prop) from rgb1,
    where prop should be a proportion in the interval 0.0 - 1.0.
    """
    return tuple(map(lambda c1, c2: int(round(c1 + (c2 - c1) * prop)), rgb1, rgb2))


def random_color():
    """
    Returns a random color drawn uniformly from the whole rgb-cube.
    """
    return rgb_color(random.random(), random.random(), random.random())


def random_hsl_color_func(hue=False, sat=False, light=False):
    """
    Returns a function that generates random colors within certain intervals.
    Each of the parameters hue, sat, and light can be either False, a constant,
    or an interval. If it is False (or not given at all), the parameter is
    randomized throughout its range, if it is an interval it is randomly drawn
    within this range, and if it is a constant it is set to that value.
    With this function you can thus construct a wide variety of random color
    generating functions.
    """

    def isnum(x):
        return type(x) in [float, int]

    def lightexp(y, x):
        return 1 - (1 - y) ** x if y > 0 else (1 + y) ** x - 1 if y < 0 else 0.0

    if not isnum(light):
        le = 1 + (1 if not isnum(sat) else 0) + (1 if not isnum(hue) else 0)
        if type(light) in [tuple, list]:
            l0 = lightexp(light[0], le)
            ld = lightexp(light[1], le) - l0
        else:
            l0 = -1.0
            ld = 2.0
    else:
        l0 = light
        ld = 0.0
    if not isnum(sat):
        se = 1 + (1 if not isnum(hue) else 0)
        if type(sat) in [tuple, list]:
            s0 = sat[0] ** se
            sd = sat[1] ** se - s0
        else:
            s0 = 0.0
            sd = 1.0
    else:
        s0 = sat
        sd = 0.0
    if not isnum(hue):
        if type(hue) in [tuple, list]:
            h0 = hue[0]
            hd = (hue[1] - hue[0]) % 1.0
        else:
            h0 = 0.0
            hd = 1.0
    else:
        h0 = hue
        hd = 0.0

    def func(*args):
        light = lightexp(l0 + random.random() * ld, 1.0 / le) if ld != 0.0 else l0
        sat = (s0 + random.random() * sd) ** (1.0 / se) if sd != 0.0 else s0
        hue = (h0 + random.random() * hd) % 1.0 if hd != 0.0 else h0
        return hsl_color(hue, sat, light)

    return func


def sprinkle_pattern(ctr, pat, rgblst, freq):
    """
    Returns a copy of pat where a random number of pixels (with freq as
    expected number) are changed to a randomly picked color from rgblst.
    """
    pat = ctr.copy_pattern(pat)
    n = randompoisson(freq)
    inds = random.sample(range(ctr.num_leds), n)
    for i in inds:
        ctr.modify_pattern(pat, i, random.choice(rgblst))
    return pat


# Example functions for creating patterns


def make_alternating_color_pattern(ctr, rgblst):
    """
    Return a pattern of alternating colors from rgblst.
    """
    n = len(rgblst)
    return ctr.make_func_pattern(lambda i: rgblst[i % n])


def make_color_spectrum_pattern(ctr, offset=0, lightness=0.0):
    """
    Return a pattern of the color spectrum along the string.
    """
    return ctr.make_func_pattern(
        lambda i: hsl_color(((i - offset) / float(ctr.num_leds)) % 1.0, 1.0, lightness),
        circular=True,
    )


def make_random_select_color_pattern(ctr, rgblst, prop=False):
    """
    Return a pattern of randomly selected colors from rgblst, optionally with
    the probabilities given by prop.
    """
    if prop and len(prop) == len(rgblst):
        return ctr.make_func_pattern(lambda i: rgblst[randomdiscrete(prop)])
    else:
        return ctr.make_func_pattern(
            lambda i: rgblst[random.randint(0, len(rgblst) - 1)]
        )


def make_random_blend_color_pattern(ctr, rgb1, rgb2):
    """
    Return a pattern of random blends of the two given colors.
    """
    return ctr.make_func_pattern(lambda i: blendcolors(rgb1, rgb2, random.random()))


def make_random_colors_pattern(ctr, lightness=0.0):
    """
    Return a pattern of randomly drawn hues of the same lightness.
    """
    return ctr.make_func_pattern(lambda i: hsl_color(random.random(), 1.0, lightness))


def make_random_lightness_pattern(ctr, hue):
    """
    Return a pattern with the same hue but randomly drawn lightnesses.
    """
    return ctr.make_func_pattern(
        lambda i: hsl_color(hue, 1.0, random.random() * 2 - 1.0)
    )


def make_random_hsl_pattern(ctr, hue=False, sat=False, light=False):
    """
    Return a pattern with random colors in the ranges specified by hue, sat, and light.
    """
    return ctr.make_func_pattern(random_hsl_color_func(hue, sat, light))



# -*- coding: utf-8 -*-

"""
xled_plus.highcontrol
~~~~~~~~~~~~~~~~~~~~~

High level control interface to the Twinkly led lights.

"""

from __future__ import absolute_import

import collections
import io
import logging
import struct
import binascii
import time
import uuid
import datetime
import math as m
from operator import xor

from xled.control import ControlInterface
from xled.util import date_from_seconds_after_midnight
from xled.security import sha1sum
from xled.exceptions import HighInterfaceError

log = logging.getLogger(__name__)

#: Time format as defined by C standard
TIME_FORMAT = "%H:%M:%S"
SHORT_TIME_FORMAT = "%H:%M"


# Should be in util.py
def seconds_after_midnight_from_string(timestr, form):
    dt = datetime.datetime.strptime(timestr, form)
    return dt.hour * 3600 + dt.minute * 60 + dt.second


class HighControlInterface(ControlInterface):
    """
    High level interface to control specific device
    """

    def __init__(self, host, hw_address=None):
        super(HighControlInterface, self).__init__(host, hw_address)
        info = self.get_device_info()
        self.num_leds = info["number_of_led"]
        self.family = info["fw_family"] if "fw_family" in info else "D"
        self.led_bytes = info["bytes_per_led"] if "bytes_per_led" in info else 3
        self.led_profile = info["led_profile"] if "led_profile" in info else "RGB"
        self.version = tuple(map(int, self.firmware_version()["version"].split(".")))
        self.string_config = self.get_led_config()["strings"]
        if not self.hw_address:
            self.hw_address = info["mac"]
        self.layout = False
        self.layout_bounds = False
        self.last_mode = None
        self.last_rt_time = 0
        self.curr_mode = self.get_mode()["mode"]

    def firmware_num_stages(self):
        if self.family == "D":
            return 2
        else:
            return 1

    def update_firmware(self, stage0, stage1=None):
        """
        Uploads firmware and runs update

        :param stage0: file-like seekable object pointing to stage0 of firmware.
        :param stage1: file-like seekable object pointing to stage1 of firmware,
                       or None if there is no stage1.
        :raises ApplicationError: on application error
        :raises HighInterfaceError: on error during update
        """
        twostage = self.fimware_num_stages() == 2
        if twostage:
            assert stage1
        else:
            assert stage1 is None
        fw_stage_sums = [None, None]
        fw_images = [stage0, stage1]
        fw_funcalls = [self.firmware_0_update, self.firmware_1_update]
        stages = [0, 1] if twostage else [0]
        for stage in stages:
            fw_images[stage].seek(0)
            fw_stage_sums[stage] = sha1sum(fw_images[stage])
            log.debug("Firmware stage %d SHA1SUM: %r", stage, fw_stage_sums[stage])
            if not fw_stage_sums[stage]:
                msg = "Failed to compute SHA1SUM for firmware stage {}.".format(stage)
                raise HighInterfaceError(msg)
                assert False

        uploaded_stage_sums = [None, None]
        for stage in stages:
            log.debug("Uploading firmware stage %d...", stage)
            fw_images[stage].seek(0)
            response = fw_funcalls[stage](fw_images[stage])
            log.debug("Firmware stage %d uploaded.", stage)
            if not response.ok:
                msg = "Failed to upload stage {}: {}".format(
                    stage, response.status_code
                )
                raise HighInterfaceError(msg)
                assert False

            uploaded_stage_sums[stage] = response.get("sha1sum")
            log.debug(
                "Uploaded stage %d SHA1SUM: %r", stage, uploaded_stage_sums[stage]
            )
            if not uploaded_stage_sums[stage]:
                msg = "Device didn't return SHA1SUM for stage {}.".format(stage)
                raise HighInterfaceError(msg)
                assert False

        if fw_stage_sums != uploaded_stage_sums:
            log.error("Firmware SHA1SUMs: %r != uploaded SHA1SUMs", fw_stage_sums)
            msg = "Firmware SHA1SUMs doesn't match uploaded SHA1SUMs."
            raise HighInterfaceError(msg)
            assert False
        else:
            log.debug("Firmware SHA1SUMs matches.")

        response = self.firmware_update(fw_stage_sums[0], fw_stage_sums[1])
        if not response.ok:
            msg = "Failed to update firmware: {}.".format(response.status_code)
            raise HighInterfaceError(msg)
            assert False

    def disable_timer(self):
        """
        Disables timer
        """
        return self.set_timer(-1, -1)

    def get_formatted_timer(self):
        """
        Gets current time and timer

        :return: namedtuple of formatted entries: current time, turn on time,
            turn off time.
        :rtype: namedtuple
        """
        Timer = collections.namedtuple("Timer", ["now", "on", "off"])

        device_response = self.get_timer()

        if xor(device_response["time_on"] == -1, device_response["time_off"] == -1):
            msg = "Inconsistent timer configuration. On: {timer_on}, off: {timer_off}".format(
                **device_response
            )
            raise HighInterfaceError(msg)

        now = device_response["time_now"]
        now_formatted = date_from_seconds_after_midnight(now).strftime(TIME_FORMAT)

        if device_response["time_on"] == -1 and device_response["time_off"] == -1:
            return Timer(now_formatted, False, False)

        on = device_response["time_on"]
        on_formatted = date_from_seconds_after_midnight(on).strftime(TIME_FORMAT)

        off = device_response["time_off"]
        off_formatted = date_from_seconds_after_midnight(off).strftime(TIME_FORMAT)

        return Timer(now_formatted, on_formatted, off_formatted)

    def set_formatted_timer(self, timestr_on, timestr_off):
        """
        Sets timer on and off times, given as strings in H:M:S or H:M format

        :param str timestr_on: time to turn on
        :param str timestr_off: time to turn off
        """
        try:
            time_on = seconds_after_midnight_from_string(timestr_on, TIME_FORMAT)
        except ValueError:
            time_on = seconds_after_midnight_from_string(timestr_on, SHORT_TIME_FORMAT)
        try:
            time_off = seconds_after_midnight_from_string(timestr_off, TIME_FORMAT)
        except ValueError:
            time_off = seconds_after_midnight_from_string(
                timestr_off, SHORT_TIME_FORMAT
            )

        return self.set_timer(time_on, time_off)

    def turn_on(self):
        """
        Turns on the device.

        Sets the mode to the last used mode before turn_off().
        If the last mode is not known, sets 'playlist' mode if there is a
        playlist, otherwise 'movie' mode if there is any uploaded movie,
        else 'effect' mode.
        """
        if self.last_mode:
            return self.set_mode(self.last_mode)
        else:
            if self.family == "D" or self.version < (2, 5, 6):
                mlst = self.get_led_movie_config()["frames_number"]
                return self.set_mode("effect" if not mlst else "movie")
            else:
                mlst = self.get_movies()["movies"]
                plst = self.get_playlist()["entries"]
                return self.set_mode("effect" if not mlst else "playlist" if plst else "movie")

    def turn_off(self):
        """
        Turns off the device.

        Remembers the previous mode, so that turn_on() can return to it.
        """
        mode = self.get_mode()["mode"]
        if mode != "off" and mode != "rt":
            self.last_mode = mode
        return self.set_mode("off")

    def is_on(self):
        """
        Returns True if device is on
        """
        return self.get_mode()["mode"] != "off"

    def set_mode(self, mode):
        """
        Sets new LED operation mode.

        :param str mode: Mode to set. One of 'movie', 'playlist', 'rt', 'demo', 'effect', 'color' or 'off'.
        This function is a wrapper around the low-level ControlInterface.set_mode,
        to remember the currently used mode.
        """
        assert mode in ("movie", "playlist", "rt", "demo", "effect", "color", "off")
        self.curr_mode = mode
        if mode != "off" and mode != "rt":
            self.last_mode = mode
        if mode == "rt":
            self.last_rt_time = time.time()
        super(HighControlInterface, self).set_mode(mode)

    # Functions for selecting what to show

    def show_movie(self, movie_or_id, fps=None):
        """
        Either starts playing an already uploaded movie with the provided id,
        or uploads a new movie and starts playing it at the provided frames-per-second.
        Note: if the movie do not fit in the remaining capacity, the old movie list is cleared.
        Switches to movie mode if necessary.
        The movie is an object suitable created with to_movie or make_func_movie.

        :param movie_or_id: either an integer id or a file-like object that points to movie
        :param fps: frames per second, or None if a movie id is given
        """
        if self.family == "D" or self.version < (2, 5, 6):
            if isinstance(movie_or_id, int) and fps is None:
                if movie_or_id != 0:
                    return False
            else:
                assert fps
                movie = movie_or_id
                numframes = movie.seek(0, 2) // (self.led_bytes * self.num_leds)
                movie.seek(0)
                self.set_led_movie_config(1000 // max(1, fps), numframes, self.num_leds)
                self.set_led_movie_full(movie)
        else:
            if isinstance(movie_or_id, int) and fps is None:
                movies = self.get_movies()["movies"]
                if movie_or_id in [entry["id"] for entry in movies]:
                    self.set_movies_current(movie_or_id)
                else:
                    return False
            else:
                assert fps
                movie = movie_or_id
                numframes = movie.seek(0, 2) // (self.led_bytes * self.num_leds)
                movie.seek(0)
                res = self.get_movies()
                capacity = res["available_frames"] - 1
                if numframes > capacity or len(res["movies"]) > 15:
                    if self.curr_mode == "movie" or self.curr_mode == "playlist":
                        self.set_mode("off")
                    self.delete_movies()
                self.set_movies_new(
                    "",
                    str(uuid.uuid4()),
                    self.led_profile.lower() + "_raw",
                    self.num_leds,
                    numframes,
                    fps,
                )
                self.set_movies_full(movie)
        if self.curr_mode != "movie":
            self.set_mode("movie")
        return True

    def upload_movie(self, movie, fps, force=False):
        """
        Uploads a new movie with the provided frames-per-second.
        Note: if the movie does not fit in the remaining capacity, and force is
        not set to True, the function just returns False, in which case the user
        can try clear_movies first.
        Does not switch to movie mode, use show_movie instead for that.
        The movie is an object suitable created with to_movie or make_func_movie.
        Returns the new movie id, which can be used in calls to show_movie or
        show_playlist.

        :param movie: a file-like object that points to movie
        :param fps: frames per second, or None if a movie id is given
        :param bool force: if remaining capacity is too low, previous movies will be removed
        :rtype: int
        """
        numframes = movie.seek(0, 2) // (self.led_bytes * self.num_leds)
        movie.seek(0)
        if self.family == "D" or self.version < (2, 5, 6):
            self.set_led_movie_config(1000 // fps, numframes, self.num_leds)
            self.set_led_movie_full(movie)
            return 0
        else:
            res = self.get_movies()
            capacity = res["available_frames"] - 1
            if numframes > capacity or len(res["movies"]) > 15:
                if force:
                    if self.curr_mode == "movie" or self.curr_mode == "playlist":
                        self.set_mode("effect")
                    self.delete_movies()
                else:
                    return False
            if self.curr_mode == "movie":
                oldid = self.get_movies_current()["id"]
            res = self.set_movies_new(
                "",
                str(uuid.uuid4()),
                self.led_profile.lower() + "_raw",
                self.num_leds,
                numframes,
                fps,
            )
            self.set_movies_full(movie)
            if self.curr_mode == "movie":
                self.set_movies_current(oldid)  # Dont change currently shown movie
            return res["id"]

    def show_pattern(self, pat):
        """
        Uploads a single pattern as a static movie, and shows it.
        Switches to movie mode if necessary.
        The parameter is a pattern object eg created with make_solid_pattern or make_func_pattern.

        :param pat: list of byte strings representing a single frame pattern
        """
        self.show_movie(self.to_movie(pat), 1)

    def show_playlist(self, lst_or_id, duration=None):
        """
        Either switches to the movie with the given id in the playlist,
        or uploads a new playlist in the form of a list where each entry is
        either an id, or a tuple with an id and a duration. The optional
        parameter duration is used for those entries without a duration.
        Switches to playlist mode if necessary.

        :param lst_or_id: integer movie id, or list of ids and durations
        :param duration: default duration to use for entries without duration
        """
        if self.family == "D" or self.version < (2, 5, 6):
            return False
        else:
            if isinstance(lst_or_id, int) and duration is None:
                plist = self.get_playlist()["entries"]
                if lst_or_id in [entry["id"] for entry in plist]:
                    if self.curr_mode != "playlist":
                        self.set_mode("playlist")
                    self.set_playlist_current(lst_or_id)
                else:
                    return False
            else:
                assert isinstance(lst_or_id, list)
                mlist = self.get_movies()["movies"]
                mdict = {entry["id"]: entry["unique_id"] for entry in mlist}
                plist = []
                for ele in lst_or_id:
                    if isinstance(ele, int):
                        plist.append(
                            {"unique_id": mdict[ele], "duration": duration or 60}
                        )
                    else:
                        plist.append({"unique_id": mdict[ele[0]], "duration": ele[1]})
                self.set_playlist(plist)
                if self.curr_mode != "playlist":
                    self.set_mode("playlist")
            return True

    def show_rt_frame(self, frame):
        """
        Uploads a frame as the next real time frame, and shows it.
        Switches to rt mode if necessary.
        The frame is either a pattern or a one-frame movie

        :param frame: a pattern or file-like object representing the frame
        """
        if self.is_pattern(frame):
            frame = self.to_movie(frame)
        if self.curr_mode != "rt" or self.last_rt_time + 50.0 < time.time():
            self.set_mode("rt")
        else:
            self.last_rt_time = time.time()
        frame.seek(0)
        # self.set_rt_frame_rest(frame)
        if self.family == "D":
            self.set_rt_frame_socket(frame, 1, self.num_leds)
        elif self.version < (2, 4, 14):
            self.set_rt_frame_socket(frame, 2)
        else:
            self.set_rt_frame_socket(frame, 3)

    def show_effect(self, effect_id):
        """
        Shows the builtin effect with the provided id.
        Switches to effect mode if necessary.

        :param int effect_id: The effect id to show
        """
        self.set_led_effects_current(effect_id)
        if self.curr_mode != "effect":
            self.set_mode("effect")

    def show_demo(self, effect_id=None):
        """
        Switches to demo mode if not there already.
        Starts from the optional provided effect id.

        :param effect_id: The optional effect id to start demo from
        """
        if effect_id:
            self.set_led_effects_current(effect_id)
        if self.curr_mode != "demo":
            self.set_mode("demo")

    def show_color(self, rgb):
        """
        Shows the given rgb color on all leds.
        Switches to color mode if necessary.

        :param tuple rgb: Tuple representing the red, green, and blue components
        """
        if self.version < (2, 7, 1):
            self.show_pattern(self.make_solid_pattern(rgb))
        else:
            self.set_led_color_rgb(*rgb)
            if self.curr_mode != "color":
                self.set_mode("color")

    def clear_movies(self):
        """
        Removes all uploaded movies and any playlist.
        If the current mode is 'movie' or 'playlist' it switches mode to 'effect'
        """
        if self.curr_mode == "movie" or self.curr_mode == "playlist":
            self.set_mode("effect")
        if self.family == "D" or self.version < (2, 5, 6):
            # No list of movies to remove in this version,
            # but disable movie mode until new movie is uploaded
            self.set_led_movie_config(1000, 0, self.num_leds)
        else:
            # The playlist is removed automatically when movies are removed
            self.delete_movies()

    # Functions for creating and manipulating movies and patterns (single frames of movies)

    def make_func_movie(self, numframes, func):
        """
        Creates a movie of a number of frames by calling a function to create each frame.
        The function is expected to take the frame index as argument and to return a
        pattern object representing the frame.

        :param int numframes: The number of frames for the movie
        :param function func: A function to produce each frame
        :rtype: _io.BytesIO
        """
        pl = []
        for i in range(numframes):
            pl.append(func(i))
        return self.to_movie(pl)

    def make_empty_movie(self):
        """
        Creates a movie of zero frames.
        Meant to be followed by several calls to add_to_movie to add frames to it.

        :rtype: _io.BytesIO
        """
        movie = io.BytesIO()
        return movie

    def is_pattern(self, pat):
        """
        Checks whether the given argument has the format of a single frame pattern.

        :param pat: object to check whether it is a pattern
        :rtype: bool
        """
        return (
            isinstance(pat, list)
            and len(pat) == self.num_leds
            and isinstance(pat[0], bytes)
        )

    def is_movie(self, movie):
        """
        Checks whether the given argument has the format of a movie.

        :param movie: object to check whether it is a movie
        :rtype: bool
        """
        return isinstance(movie, io.BytesIO)

    def add_to_movie(self, movie, pat):
        """
        Adds one pattern as a frame to the end of a movie.

        :param movie: file-like object representing the movie
        :param pat: object representing the pattern
        :rtype: _io.BytesIO
        """
        assert self.is_pattern(pat)
        movie.seek(0, 2)
        movie.write(b"".join(pat))
        movie.seek(0, 0)

    def to_movie(self, patlst):
        """
        Creates a movie from either a single pattern or a list of patterns.

        :param patlst: pattern or list of patterns
        :rtype: _io.BytesIO
        """
        movie = io.BytesIO()
        if isinstance(patlst, list):
            for ele in patlst:
                if isinstance(ele, list):
                    ele = b"".join(ele)
                movie.write(ele)
        else:
            movie.write(patlst)
        movie.seek(0)
        return movie

    def circind(self, ind):
        """
        Internal function used to fascilitate linear or circular effects. That
        is, if the device consists of two strings, flip the led indices of one
        of the strings so they start at the extreme end of the first string
        and runs into the middle where the strings meet and then continue out
        on the other string. If the extreme ends of the two strings are
        arranged to meet again, it allows for circular patterns.
        """
        if len(self.string_config) == 2 and ind < self.string_config[0]["length"]:
            return self.string_config[0]["length"] - 1 - ind
        else:
            return ind

    def make_pixel(self, r, g, b):
        """
        Internal function to produce one pixel of a pattern from given r, g
        and b values. Handles both RGB and RGBW led profiles (for now always
        setting the white led to zero).

        :param int r: red component
        :param int g: green component
        :param int b: blue component
        :rtype: bytes
        """
        if self.led_bytes == 4:
            return struct.pack(">BBBB", 0, r, g, b)
        else:
            return struct.pack(">BBB", r, g, b)

    def make_solid_pattern(self, rgb):
        """
        Creates a one-colored pattern with the given rgb value tuple.

        :param tuple rgb: color as an rgb tuple
        :rtype: list representing the pattern
        """
        pat = [self.make_pixel(*rgb)] * self.num_leds
        return pat

    def make_func_pattern(self, func, circular=False):
        """
        Creates a pattern by calling the given function for each led.
        The function is expected to take the led index as argument and to
        return a color as an rgb tuple for that led.

        :param function func: function to return the color of each pixel
        :param bool circular: Flip the led indices on two-string devices to enable circular patterns
        :rtype: list representing the pattern
        """
        pat = [False] * self.num_leds
        for i in range(self.num_leds):
            (r, g, b) = func(i)
            if circular:
                pat[self.circind(i)] = self.make_pixel(r, g, b)
            else:
                pat[i] = self.make_pixel(r, g, b)
        return pat

    def fetch_layout(self, aspect=False):
        if self.family != 'D' and self.version > (2, 2, 1):
            res = self.get_led_layout()
            if res["source"] == "3d":
                if aspect:
                    self.layout = [
                        (p["x"] * aspect[0] * 0.5, p["y"], p["z"] * aspect[1] * 0.5)
                        for p in res["coordinates"]
                    ]
                else:
                    self.layout = [
                        (p["x"] * 0.5, p["y"], p["z"] * 0.5) for p in res["coordinates"]
                    ]
                dim = 3
            elif res["source"] == "2d":
                if aspect:
                    self.layout = [
                        (p["x"] * aspect[0] * 0.5, p["y"]) for p in res["coordinates"]
                    ]
                else:
                    self.layout = [(p["x"] * 0.5, p["y"]) for p in res["coordinates"]]
                dim = 2
            else:
                self.layout = [(p["x"],) for p in res["coordinates"]]
                dim = 1
        else:
            self.layout = [(float(self.circind(i)) / (self.num_leds - 1),) for i in range(self.num_leds)]
            dim = 1
        bounds = []
        cent = []
        rad = 0.0
        for d in range(dim):
            vals = [p[d] for p in self.layout]
            bounds.append((min(vals), max(vals)))
            cent.append(sum(vals) / len(vals))
        for p in self.layout:
            r2 = sum([(p[d] - cent[d]) ** 2 for d in range(dim)])
            if r2 > rad:
                rad = r2
        self.layout_bounds = {
            "dim": dim,
            "bounds": bounds,
            "center": cent,
            "radius": rad ** 0.5,
        }
        if dim == 3:
            crad = 0.0
            hsrad = 0.0
            for p in self.layout:
                r2 = (p[0] - cent[0]) ** 2 + (p[2] - cent[2]) ** 2
                if r2 > crad:
                    crad = r2
                r2 += (p[1] - bounds[1][0]) ** 2
                if r2 > hsrad:
                    hsrad = r2
            self.layout_bounds["cylradius"] = crad ** 0.5
            self.layout_bounds["halfsradius"] = hsrad ** 0.5

    def layout_transform(self, pos, style):
        # style == 'square', 'rect', 'centered', 'cylinder', 'sphere'
        if style == "square":
            # Stretch everything into [0, 1] in each coordinate
            return tuple(
                (v - b[0]) / (b[1] - b[0])
                for v, b in zip(pos, self.layout_bounds["bounds"])
            )
        elif style == "rect":
            # Keep aspect ratio, largest into [-1,1]
            cent = ((b[0] + b[1]) / 2 for b in self.layout_bounds["bounds"])
            width = max((b[1] - b[0]) / 2 for b in self.layout_bounds["bounds"])
            return tuple((v - c) / width for v, c in zip(pos, cent))
        elif style == "centered":
            # Origo in center, max radius 1.0
            rad = self.layout_bounds["radius"]
            return tuple(
                (v - c) / rad for v, c in zip(pos, self.layout_bounds["center"])
            )
        elif style == "cylinder" and self.layout_bounds["dim"] == 3:
            # xz-radius max 1, angle in [-180,180], y in [0, 1]
            crad = self.layout_bounds["cylradius"]
            ybounds = self.layout_bounds["bounds"][1]
            p = tuple((v - c) / crad for v, c in zip(pos, self.layout_bounds["center"]))
            return (
                m.sqrt(p[0] ** 2 + p[2] ** 2),
                m.atan2(p[0], p[2]) * 180.0 / m.pi,
                (pos[1] - ybounds[0]) / (ybounds[1] - ybounds[0]),
            )
        elif style == "sphere" and self.layout_bounds["dim"] == 3:
            # radius max 1, longitude [-180,180], latitude [-90,90]
            rad = self.layout_bounds["radius"]
            p = tuple((v - c) / rad for v, c in zip(pos, self.layout_bounds["center"]))
            return (
                m.sqrt(p[0] ** 2 + p[1] ** 2 + p[2] ** 2),
                m.atan2(p[0], p[2]) * 180.0 / m.pi,
                m.atan2(p[1], m.sqrt(p[0] ** 2 + p[2] ** 2)) * 180.0 / m.pi,
            )
        elif style == "halfsphere" and self.layout_bounds["dim"] == 3:
            # radius max 1, longitude [-180,180], polar angle [0,90]
            rad = self.layout_bounds["halfsradius"]
            cent = (self.layout_bounds["center"][0], self.layout_bounds["bounds"][1][0], self.layout_bounds["center"][2])
            p = tuple((v - c) / rad for v, c in zip(pos, cent))
            return (
                m.sqrt(p[0] ** 2 + p[1] ** 2 + p[2] ** 2),
                m.atan2(p[0], p[2]) * 180.0 / m.pi,
                m.atan2(m.sqrt(p[0] ** 2 + p[2] ** 2), p[1]) * 180.0 / m.pi,
            )
        else:
            return pos

    def make_layout_pattern(self, func, style=None, index=False):
        """
        Creates a pattern by calling the given function for each led.
        The function is expected to take the led physical position as
        argument (1d, 2d, or 3d depending on the layout source) and to
        return a color as an rgb tuple for that led.

        :param function func: function to return the color of each pixel
        :rtype: list representing the pattern
        """
        if not self.layout:
            self.fetch_layout()
        pat = [False] * self.num_leds
        for i in range(self.num_leds):
            pos = self.layout_transform(self.layout[i], style)
            if index:
                (r, g, b) = func(pos, i)
            else:
                (r, g, b) = func(pos)
            pat[i] = self.make_pixel(r, g, b)
        return pat

    def adjust_layout_aspect(self, aspect_xy, aspect_zy=False):
        if aspect_xy and aspect_zy:
            self.fetch_layout((aspect_xy, aspect_zy))
        elif aspect_xy:
            self.fetch_layout((aspect_xy, aspect_xy))
        else:
            self.fetch_layout()

    def get_layout_bounds(self):
        if not self.layout:
            self.fetch_layout()
        return self.layout_bounds

    def copy_pattern(self, pat):
        """
        Make a copy of a pattern.
        In case you want to make destructive operations on one of them.

        :param pat: object representing the pattern
        :rtype: list representing the pattern
        """
        return [ele for ele in pat]

    def modify_pattern(self, pat, ind, rgb, circular=False):
        """
        Modifies one pixel in a pattern.
        Changes the pattern in place. Make sure to copy it if you need the old one.

        :param pat: object representing the pattern
        :param int ind: led index in the pattern
        :param tuple rgb: color as an rgb tuple
        :param bool circular: Flip the led indices on two-string devices to enable circular patterns
        :rtype: list representing the pattern (the same object as pat)
        """
        if circular:
            pat[self.circind(ind)] = self.make_pixel(*rgb)
        else:
            pat[ind] = self.make_pixel(*rgb)
        return pat

    def shift_pattern(self, pat, step, rgb, circular=False):
        """
        Shifts the pattern a number of steps, padding with the provided rgb color.
        Non-destructive, leaving the original pattern unmodified.

        :param pat: object representing the pattern
        :param int step: steps to shift, can be positive or negative
        :param tuple rgb: color as an rgb tuple
        :param bool circular: Flip the led indices on two-string devices to enable circular patterns
        :rtype: list representing the pattern
        """
        pix = self.make_pixel(*rgb)
        if circular and len(self.string_config) == 2:
            n1 = self.string_config[0]["length"]
            n2 = self.num_leds
            p1 = pat[0:n1]
            p2 = pat[n1:n2]
            if step > 0:
                for i in range(step):
                    p2 = p1[:1] + p2[:-1]
                    p1 = p1[1:] + [pix]
            else:
                for i in range(-step):
                    p1 = p2[:1] + p1[:-1]
                    p2 = p2[1:] + [pix]
            pat = p1 + p2
        else:
            if step > 0:
                pat = [pix] * step + pat[:-step]
            else:
                pat = pat[-step:] + [pix] * -step
        return pat

    def rotate_pattern(self, pat, step, circular=False):
        """
        Shifts the pattern a number of steps with rotation, so that pixels
        shifted out at one end emerges at the other end.
        Non-destructive, leaving the original pattern unmodified.

        :param pat: object representing the pattern
        :param int step: steps to shift, can be positive or negative
        :param bool circular: Flip the led indices on two-string devices to enable circular patterns
        :rtype: list representing the pattern
        """
        if circular and len(self.string_config) == 2:
            n1 = self.string_config[0]["length"]
            n2 = self.num_leds
            p1 = pat[0:n1]
            p2 = pat[n1:n2]
            if step > 0:
                for i in range(step):
                    tmp = p1[0]
                    p1 = p1[1:] + p2[-1:]
                    p2 = [tmp] + p2[:-1]
            else:
                for i in range(-step):
                    tmp = p1[-1]
                    p1 = p2[:1] + p1[:-1]
                    p2 = p2[1:] + [tmp]
            pat = p1 + p2
        else:
            pat = pat[-step:] + pat[:-step]
        return pat

    def permute_pattern(self, pat, perm, circular=False):
        """
        Permutes the pattern according to the provided permutation list.
        The new index 'i' will get the same color as the old index 'perm[i]'.
        Non-destructive, leaving the original pattern unmodified.

        :param pat: object representing the pattern
        :param list perm: permutation list of source indices
        :param bool circular: Flip the led indices on two-string devices to enable circular patterns
        :rtype: list representing the pattern
        """
        newpat = [False] * len(pat)
        if circular:
            for i, k in enumerate(perm):
                newpat[self.circind(i)] = pat[self.circind(k)]
        else:
            for i, k in enumerate(perm):
                newpat[i] = pat[k]
        return newpat

    def save_movie(self, name, movie, fps):
        """
        Save the movie object on file.
        The movie file is text based and starts with a header containing
        the number of frames, number of leds, number of bytes per led, and
        the suggested frames per second. After the header follows one line per
        frame as a hexadecimal string. This format makes it easier to share
        movies between different devices and even different led profiles.
        """
        bytesperframe = self.led_bytes * self.num_leds
        numframes = movie.seek(0, 2) // bytesperframe
        movie.seek(0)
        f = open(name, "w")
        f.write("{} {} {} {}\n".format(numframes, self.num_leds, self.led_bytes, fps))
        for i in range(numframes):
            f.write(binascii.hexlify(movie.read(bytesperframe)).decode() + "\n")
        f.close()

    def load_movie(self, name):
        """
        Read a movie from a file (produced by save_movie).
        Returns both the movie object and the suggested frames-per-second in a tuple.
        Some effort is made to convert movies between different devices:
        If the number of leds are different, each frame is padded or truncated
        at both ends. If the led profile is different, the white component is
        removed or added (as zero).
        """
        f = open(name, "r")
        head = list(map(int, f.readline().strip("\n").split(" ")))
        numframes = head[0]
        fps = head[3]
        movie = io.BytesIO()
        if head[1] == self.num_leds and head[2] == self.led_bytes:
            for i in range(numframes):
                movie.write(binascii.unhexlify(f.readline().strip("\n")))
        else:
            for i in range(numframes):
                s = binascii.unhexlify(f.readline().strip("\n"))
                if head[2] == 3 and self.led_bytes == 4:
                    s = b"".join(
                        [chr(0) + s[3 * i : 3 * i + 3] for i in range(len(s) // 3)]
                    )
                elif head[2] == 4 and self.led_bytes == 3:
                    s = b"".join([s[4 * i + 1 : 4 * i + 4] for i in range(len(s) // 4)])
                if head[1] < self.num_leds:
                    diff = self.num_leds - head[1]
                    s = (
                        chr(0) * (diff // 2 * self.led_bytes)
                        + s
                        + chr(0) * ((diff - diff // 2) * self.led_bytes)
                    )
                elif head[1] > self.num_leds:
                    hdiff = (head[1] - self.num_leds) // 2
                    s = s[hdiff : hdiff + self.num_leds * self.led_bytes]
                movie.write(s)
        movie.seek(0)
        return (movie, fps)



from .sample_setup import *

ctr = setup_control()
oldmode = ctr.get_mode()["mode"]
eff = ColorMeanderEffect(ctr, "solid")
eff.launch_rt()
print("Started continuous effect - press Return to stop it")
input()
eff.stop_rt()
ctr.set_mode(oldmode)



"""
Day 12 - Gradients of blue

Gradients of blue, green and purple sweeping over the lights in rotating direction.
Uses the layout, and for best effect you may have to adjust the aspect ratio:
The parameter to adjust_layout_aspect should be set to width/height of the lights.
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
ctr.adjust_layout_aspect(1.0)  # How many times wider than high is the led installation?
cols = [hsl_color(0.0, 0.0, 0.0),hsl_color(0.02, 1.0, 0.0), hsl_color(0.9, 1.0, 0.0), hsl_color(0.0, 0.0, 0.0), hsl_color(0.06, 1.0, 0.0), hsl_color(0.18, 1.0, 0.0)]
eff = RotatingAngleGradientSequence(ctr, cols, speed=0.04, folds=0.5, angle=-45, torque=7.2)
eff.preferred_frames = 400
eff.launch_movie()



"""
xled_plus.effect_base
~~~~~~~~~~~~~~~~~

Author: Anders Holst (anders.holst@ri.se), 2021

Base class Effect to provide a consistent interface to moving effects.
As outer API, i.e to users of an effect, it provides these functions:
'launch_movie()' to create a movie of the effect and upload and start playing it.
'save_movie()' to create a movie of the effect and save it to file for later use.
'launch_rt()' for playing the effect in real time.
'stop_rt()' for stopping the currently played real time effect.

As inner API, i.e for communicating with its subclasses, it requires each subclass
to provide two functions:
'reset(numframes)' to prepare all inner data structures to start generating an effect.
'getnext()' to produce and return the next frame pattern of the effect.
The Effect class also provides two member variables that can be set by the subclass:
'preferred_frames' and 'preferred_fps', as a sitable number of frames in a movie
of the effect, and the suggested number of frames per second of the effect.
Note that the 'reset(numframes)' function will either get the number of frames
required for the movie (then typically the same as 'preferred_frames') or False
if a real time effect is requested. If there is a non-False numframes, 'reset'
should try to set up data structures to make sure that after this many frames
the movie will seamlessly return to the first frame.
"""

import sys
import time

if sys.version_info.major == 2:
    from threading import _Timer

    TimerX = _Timer
else:
    from threading import Timer

    TimerX = Timer


class RepeatedTimer(TimerX):
    def run(self):
        lasttime = time.time()
        self.function(*self.args, **self.kwargs)
        while not self.finished.wait(
            max(0.0, self.interval - (time.time() - lasttime))
        ):
            lasttime = time.time()
            self.function(*self.args, **self.kwargs)


effect_timer = None


class Effect(object):
    def __init__(self, ctr):
        self.ctr = ctr
        self.preferred_frames = 120
        self.preferred_fps = 8

    def reset(self, numframes=False):
        pass  # provided by subclass

    def getnext(self):
        pass  # provided by subclass

    def launch_rt(self):
        global effect_timer

        def doit():
            self.ctr.show_rt_frame(self.getnext())

        if effect_timer:
            effect_timer.cancel()
        effect_timer = RepeatedTimer(1.0 / self.preferred_fps, doit)
        self.reset(False)
        effect_timer.start()
        return True

    def stop_rt(self):
        global effect_timer
        if effect_timer:
            effect_timer.cancel()
        effect_timer = None

    def make_movie(self, numframes):
        frames = []
        self.reset(numframes)
        for i in range(numframes):
            frames.append(self.getnext())
        return self.ctr.to_movie(frames)

    def launch_movie(self):
        self.stop_rt()
        self.ctr.show_movie(self.make_movie(self.preferred_frames), self.preferred_fps)

    def save_movie(self, name):
        self.ctr.save_movie(
            name, self.make_movie(self.preferred_frames), self.preferred_fps
        )


def stop_rt():
    global effect_timer
    if effect_timer:
        effect_timer.cancel()
        effect_timer = None



from xled.discover import discover
from xled_plus.highcontrol import HighControlInterface
from xled_plus.ledcolor import *
from xled_plus.pattern import *
from xled_plus.effects import *
from xled_plus.sequence import *
from xled_plus.shapes import *
from sys import argv

def setup_control():
    if len(argv) > 1:
        host = argv[1]
    else:
        dev = discover()
        host = dev.ip_address
    return HighControlInterface(host)





"""
Day 14 - Looping lights, red, magenta, green

Another variant of SparkleEffect, this time each led "looping" from black
through one of the selected colors to white, and then dims back to black
again (through gray). Gives nice white mix among the pulsating colors.
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
nsparks = 0.4 * ctr.num_leds / 32.0
colfunc = selected_color_func([(0.6, 1.0), (0.7, 1.0), (0.28, 1.0)])
SparkleEffect(ctr, nsparks, colfunc, looplight_func(16, 16)).launch_movie()



"""
xled_plus.shapes
~~~~~~~~~~~~~~~~

Author: Anders Holst (anders.holst@ri.se), 2021

This module contains classes for creating scenes of moving (or still) objects.
Objects implemented are various geometrical shapes, such as polygons, stars,
and circles, but also simple letters and numbers. 
"""


import math as m
from random import random, gauss, uniform
from copy import copy

from xled_plus.ledcolor import hsl_color
from xled_plus.effect_base import Effect
from xled_plus.colormeander import ColorMeander
from xled_plus.pattern import random_hsl_color_func, dimcolor



class Scene(Effect):
    def __init__(self, ctr):
        super(Scene, self).__init__(ctr)
        self.shapes = []
        self.occvec = [False] * ctr.num_leds
        self.bgfunc = False
        self.proj2D3D = False  # 'cylshell', 'cylbase', 'halfsphere' 

    def add_shape(self, sh):
        self.shapes.append(sh)
        self.shapes.sort(key=lambda sh: sh.depth)

    def remove_shape(self, sh):
        self.shapes.remove(sh)

    def update(self, step):
        for sh in self.shapes:
            sh.update(step)

    def blend_colors(self, colors):
        if colors:
            return tuple(map(lambda *args: int(round(sum(args) / len(args))), *colors))
        else:
            return False

    def get_color(self, coord, ind):
        goaldepth = False
        colors = []
        for sh in self.shapes:
            if goaldepth and sh.depth != goaldepth:
                break
            else:
                col = sh.get_color(coord, ind)
                if col:
                    if not goaldepth:
                        goaldepth = sh.depth
                    colors.append(col)
        self.occvec[ind] = True if colors else False
        return self.blend_colors(colors) or (self.bgfunc(coord, ind) if self.bgfunc else (0, 0, 0))

    def reset(self, numframes):
        pass

    def getnext(self):
        self.update(1)
        dim = self.ctr.get_layout_bounds()["dim"]
        if dim == 3 and self.proj2D3D:
            if self.proj2D3D == "cylbase":
                fact = self.ctr.get_layout_bounds()["radius"] / self.ctr.get_layout_bounds()["cylradius"]
                colfunc = lambda pos, ind: self.get_color((pos[0]*fact, pos[2]*fact), ind)
                return self.ctr.make_layout_pattern(colfunc, style="centered", index=True)
            elif self.proj2D3D == "cylshell":
                hyp = ((m.pi * self.ctr.get_layout_bounds()["cylradius"]) ** 2 + 0.25) ** 0.5
                xfact = m.pi / 180.0 * self.ctr.get_layout_bounds()["cylradius"] / hyp
                yfact = 1.0 / hyp
                colfunc = lambda pos, ind: self.get_color((pos[1]*xfact, (pos[2] - 0.5)*yfact), ind)
                return self.ctr.make_layout_pattern(colfunc, style="cylinder", index=True)
            elif self.proj2D3D == "halfsphere":
                colfunc = lambda pos, ind: self.get_color((m.sin(pos[1]*m.pi/180.0)*pos[2]/90.0, -m.cos(pos[1]*m.pi/180.0)*pos[2]/90.0), ind)
                return self.ctr.make_layout_pattern(colfunc, style="halfsphere", index=True)
        return self.ctr.make_layout_pattern(self.get_color, style="centered", index=True)

    def getoccupancy(self):
        return self.occvec

    def get_scene_bounds(self):
        bounds = self.ctr.get_layout_bounds()
        if bounds["dim"] == 3:
            if self.proj2D3D == "cylbase":
                fact = self.ctr.get_layout_bounds()["radius"] / self.ctr.get_layout_bounds()["cylradius"]
                return [(fact * bounds["bounds"][0][0], fact * bounds["bounds"][0][1]),
                        (fact * bounds["bounds"][2][0], fact * bounds["bounds"][2][1])]
            elif self.proj2D3D == "cylshell":
                hyp = ((m.pi * bounds["cylradius"]) ** 2 + 0.25) ** 0.5
                xfact = m.pi * bounds["cylradius"] / hyp
                yfact = 0.5 / hyp
                return [(-xfact, xfact), (-yfact, yfact)]
            elif self.proj2D3D == "halfsphere":
                return [(-1.0, 1.0), (-1.0, 1.0)]
            else:
                return bounds["bounds"]
        elif bounds["dim"] == 2:
            fact = 1.0 / bounds["radius"]
            return [(fact * bounds["bounds"][0][0], fact * bounds["bounds"][0][1]),
                    (fact * bounds["bounds"][1][0], fact * bounds["bounds"][1][1])]
        else:
            return bounds["bounds"]


class Shape(object):
    def __init__(self, cent, angle):
        self.depth = 0
        self.color = False
        self.cent = cent
        self.speed = (0.0, ) * len(cent)
        self.angle = angle * m.pi / 180.0
        self.torque = 0.0

    def is_inside(self, coord):
        pass

    def get_color(self, coord, ind):
        if self.is_inside(coord):
            return self.color(coord, ind) if callable(self.color) else self.color
        else:
            return False

    def set_depth(self, depth):
        self.depth = depth

    def set_speed(self, vx, vy):
        self.speed = (vx, vy)

    def set_torque(self, va):
        self.torque = va * m.pi / 180.0

    def update(self, step):
        self.cent = tuple(map(lambda c, s: c + step * s, self.cent, self.speed))
        self.angle = (self.angle + step * self.torque) % 360.0


class Blob(Shape):
    def __init__(self, cent, rad, col):
        super(Blob, self).__init__(cent, 0.0)
        self.rad = rad
        self.color = col

    def is_inside(self, coord):
        return (
            sum(map(lambda x1, x2: (x1 - x2) ** 2, self.cent, coord)) <= self.rad ** 2
        )

    def get_color(self, coord, ind):
        dist = (
            m.sqrt(sum(map(lambda x1, x2: (x1 - x2) ** 2, self.cent, coord))) / self.rad
        )
        if dist > 1.0:
            return False
        else:
            col = self.color(coord, ind) if callable(self.color) else self.color
            return dimcolor(col, 1.0 - dist)


class Polygon(Shape):
    def __init__(self, num, cent, angle, smallrad, col):
        assert num >= 3
        super(Polygon, self).__init__(cent, angle)
        self.num = num
        self.rad1 = smallrad
        self.rad2 = smallrad / m.cos(m.pi / num)
        self.color = col

    def is_inside(self, coord):
        dist = m.sqrt(sum(map(lambda x1, x2: (x1 - x2) ** 2, self.cent, coord)))
        if dist > self.rad2:
            return False
        elif dist <= self.rad1:
            return True
        else:
            dp = list(map(lambda x1, x2: (x1 - x2), self.cent, coord))
            ang = (self.angle + m.atan2(dp[0], -dp[1]) + m.pi / self.num) % (
                2 * m.pi / self.num
            ) - m.pi / self.num
            return dist <= self.rad1 / m.cos(ang)


class Ellipse(Shape):
    def __init__(self, cent, angle, largerad, smallrad, col):
        super(Ellipse, self).__init__(cent, angle)
        self.rad1 = smallrad
        self.rad2 = largerad
        self.color = col

    def is_inside(self, coord):
        dist = m.sqrt(sum(map(lambda x1, x2: (x1 - x2) ** 2, self.cent, coord)))
        if dist > self.rad2:
            return False
        elif dist <= self.rad1:
            return True
        else:
            dp = list(map(lambda x1, x2: (x1 - x2), self.cent, coord))
            ang = -self.angle + m.atan2(dp[0], dp[1])
            return dist <= self.rad1 * self.rad2 / m.sqrt(
                (self.rad1 * m.cos(ang)) ** 2 + (self.rad2 * m.sin(ang)) ** 2
            )


class Star(Shape):
    def __init__(self, num, cent, angle, largerad, smallrad, col):
        assert num >= 2
        super(Star, self).__init__(cent, angle)
        self.num = num
        self.set_radius(largerad, smallrad)
        self.color = col

    def set_radius(self, largerad, smallrad):
        self.rad1 = smallrad
        self.rad2 = largerad
        self.ang0 = m.atan(
            (m.cos(m.pi / self.num) - smallrad / largerad) / m.sin(m.pi / self.num)
        )
        self.rad0 = smallrad * m.cos(self.ang0)

    def is_inside(self, coord):
        dist = m.sqrt(sum(map(lambda x1, x2: (x1 - x2) ** 2, self.cent, coord)))
        if dist > self.rad2:
            return False
        elif dist <= self.rad1:
            return True
        else:
            dp = list(map(lambda x1, x2: (x1 - x2), self.cent, coord))
            ang = abs(
                (-self.angle + m.atan2(dp[0], dp[1])) % (2 * m.pi / self.num)
                - m.pi / self.num
            )
            return dist <= self.rad0 / m.cos(ang + self.ang0)


# Tänk ett antal punkter med linjer eller arcs mellan
# Vidare en standardstorlek och linjevidd, som kan skalas och få en riktning
#
# En arc (p1, p2, ang). centrum (0,0) start (-1,0) vinkel ang -> slutpunkt (-cos(ang), sin(ang))
# Hitta en mappning p = A + Rx så att (-1,0) -> p1 och (-c,s) -> p2
#
# R*(x,y) = (0,x^2+y^2) -> ((y, -x),(x, y))
#
class Lineart2D(Shape):
    def __init__(self):
        super(Lineart2D, self).__init__((0.0, 0.0), 0.0)
        self.points = []
        self.lines = []
        self.arcs = []
        self.extent = [0.0, 0.0, 0.0, 0.0]  # minx miny maxx maxy
        self.off = [0.0, 0.0]
        self.mat = [[1.0, 0.0], [0.0, 1.0]]
        self.lw = 0.1
        self.color = (0, 0, 0)

    def add_extent(self, minx, miny, maxx, maxy):
        if self.extent[0] > minx:
            self.extent[0] = minx
        if self.extent[1] > miny:
            self.extent[1] = miny
        if self.extent[2] < maxx:
            self.extent[2] = maxx
        if self.extent[3] < maxy:
            self.extent[3] = maxy

    def add_segment(self, p1, p2, ang):
        # If p1==p2 its a point, if ang==0 its a line, else its an arc
        # In all cases, produce a local transformation A + Rx, and increase extent
        if p1 == p2:
            self.points.append([-p1[0], -p1[1]])
            self.add_extent(p1[0], p1[1], p1[0], p1[1])
        elif ang == 0.0:
            dp = [p2[0] - p1[0], p2[1] - p1[1]]
            rad = m.sqrt(dp[0] * dp[0] + dp[1] * dp[1])
            self.lines.append(
                (
                    [-p1[0], -p1[1]],
                    [[dp[1] / rad, -dp[0] / rad], [dp[0] / rad, dp[1] / rad]],
                    rad,
                )
            )
            self.add_extent(
                min(p1[0], p2[0]),
                min(p1[1], p2[1]),
                max(p1[0], p2[0]),
                max(p1[1], p2[1]),
            )
            tmp = [-p1[0], -p1[1]]
            if tmp not in self.points:
                self.points.append(tmp)
            tmp = [-p2[0], -p2[1]]
            if tmp not in self.points:
                self.points.append(tmp)
        else:
            if ang < 0.0:
                ang = -ang
                (p1, p2) = (p2, p1)
            cot = m.tan((180.0 - ang) * m.pi / 360.0) / 2.0
            dp = [p2[0] - p1[0], p2[1] - p1[1]]
            cent = [dp[0] * 0.5 + dp[1] * cot, dp[1] * 0.5 - dp[0] * cot]
            rad = m.sqrt(cent[0] * cent[0] + cent[1] * cent[1])
            aa = ang * m.pi / 180.0 - m.pi
            self.arcs.append(
                (
                    [-p1[0] - cent[0], -p1[1] - cent[1]],
                    [[cent[0] / rad, cent[1] / rad], [-cent[1] / rad, cent[0] / rad]],
                    rad,
                    aa,
                )
            )
            self.add_extent(
                min(p1[0], p2[0])
                if m.atan2(-cent[1], -cent[0]) > aa
                else p1[0] + cent[0] - rad,
                min(p1[1], p2[1])
                if m.atan2(cent[0], -cent[1]) > aa
                else p1[1] + cent[1] - rad,
                max(p1[0], p2[0])
                if m.atan2(cent[1], cent[0]) > aa
                else p1[0] + cent[0] + rad,
                max(p1[1], p2[1])
                if m.atan2(-cent[0], cent[1]) > aa
                else p1[1] + cent[1] + rad,
            )
            tmp = [-p1[0], -p1[1]]
            if tmp not in self.points:
                self.points.append(tmp)
            tmp = [-p2[0], -p2[1]]
            if tmp not in self.points:
                self.points.append(tmp)

    def trans(self, coord, off, mat):
        return [
            (off[0] + coord[0]) * mat[0][0] + (off[1] + coord[1]) * mat[0][1],
            (off[0] + coord[0]) * mat[1][0] + (off[1] + coord[1]) * mat[1][1],
        ]

    def is_inside(self, coord):
        # global transform
        # jämför extent
        # för varje segment, lokal transform och kolla närhet
        pp = self.trans(coord, self.off, self.mat)
        if (
            pp[0] < self.extent[0] - self.lw * 0.5
            or pp[1] < self.extent[1] - self.lw * 0.5
            or pp[0] > self.extent[2] + self.lw * 0.5
            or pp[1] > self.extent[3] + self.lw * 0.5
        ):
            return False
        for seg in self.points:
            p0 = [pp[0] + seg[0], pp[1] + seg[1]]
            if p0[0] * p0[0] + p0[1] * p0[1] <= self.lw * self.lw / 4:
                return True
        for seg in self.lines:
            p0 = self.trans(pp, seg[0], seg[1])
            if (
                p0[0] <= self.lw / 2
                and p0[0] >= -self.lw / 2
                and p0[1] >= 0.0
                and p0[1] <= seg[2]
            ):
                return True
        for seg in self.arcs:
            p0 = self.trans(pp, seg[0], seg[1])
            r = m.sqrt(p0[0] * p0[0] + p0[1] * p0[1])
            a = m.atan2(-p0[1], p0[0])
            if r >= seg[2] - self.lw / 2 and r <= seg[2] + self.lw / 2 and a <= seg[3]:
                return True
        return False


letters = {
    "A": [
        ([-0.35, 0.0], [-0.05, 1.0], 0.0),
        ([0.35, 0.0], [0.05, 1.0], 0.0),
        ([-0.05, 1.0], [0.05, 1.0], 0.0),
        ([-0.2, 0.5], [0.2, 0.5], 0.0),
    ],
    "B": [
        ([-0.35, 0.0], [-0.35, 1.0], 0.0),
        ([-0.35, 1.0], [0.1, 1.0], 0.0),
        ([-0.35, 0.5], [0.1, 0.5], 0.0),
        ([-0.35, 0.0], [0.1, 0.0], 0.0),
        ([0.1, 1.0], [0.1, 0.5], 180.0),
        ([0.1, 0.5], [0.1, 0.0], 180.0),
    ],
    "C": [
        ([-0.35, 0.35], [0.247, 0.103], -135.0),
        ([-0.35, 0.65], [0.247, 0.897], 135.0),
        ([-0.35, 0.35], [-0.35, 0.65], 0.0),
    ],
    "D": [
        ([-0.35, 0.0], [-0.35, 1.0], 0.0),
        ([-0.35, 1.0], [0.0, 1.0], 0.0),
        ([-0.35, 0.0], [0.0, 0.0], 0.0),
        ([0.0, 1.0], [0.35, 0.65], 90.0),
        ([0.0, 0.0], [0.35, 0.35], -90.0),
        ([0.35, 0.25], [0.35, 0.75], 0.0),
    ],
    "E": [
        ([-0.35, 0.0], [-0.35, 1.0], 0.0),
        ([-0.35, 1.0], [0.35, 1.0], 0.0),
        ([-0.35, 0.5], [0.25, 0.5], 0.0),
        ([-0.35, 0.0], [0.35, 0.0], 0.0),
    ],
    "F": [
        ([-0.35, 0.0], [-0.35, 1.0], 0.0),
        ([-0.35, 1.0], [0.35, 1.0], 0.0),
        ([-0.35, 0.5], [0.25, 0.5], 0.0),
    ],
    "G": [
        ([-0.35, 0.35], [0.35, 0.35], -180.0),
        ([-0.35, 0.65], [0.247, 0.897], 135.0),
        ([-0.35, 0.35], [-0.35, 0.65], 0.0),
        ([0.35, 0.35], [0.35, 0.5], 0.0),
        ([0.1, 0.5], [0.35, 0.5], 0.0),
    ],
    "H": [
        ([-0.35, 0.0], [-0.35, 1.0], 0.0),
        ([0.35, 0.0], [0.35, 1.0], 0.0),
        ([-0.35, 0.5], [0.35, 0.5], 0.0),
    ],
    "I": [
        ([0.0, 0.0], [0.0, 1.0], 0.0),
        ([-0.15, 0.0], [0.15, 0.0], 0.0),
        ([-0.15, 1.0], [0.15, 1.0], 0.0),
    ],
    "J": [
        ([0.15, 0.15], [0.15, 1.0], 0.0),
        ([0.15, 0.15], [-0.15, 0.15], 180.0),
    ],
    "K": [
        ([-0.35, 0.0], [-0.35, 1.0], 0.0),
        ([-0.35, 0.5], [0.35, 1.0], 0.0),
        ([-0.35, 0.5], [0.35, 0.0], 0.0),
    ],
    "L": [
        ([-0.35, 0.0], [-0.35, 1.0], 0.0),
        ([-0.35, 0.0], [0.35, 0.0], 0.0),
    ],
    "M": [
        ([-0.35, 0.0], [-0.35, 1.0], 0.0),
        ([-0.35, 1.0], [0.0, 0.5], 0.0),
        ([0.0, 0.5], [0.35, 1.0], 0.0),
        ([0.35, 1.0], [0.35, 0.0], 0.0),
    ],
    "N": [
        ([-0.35, 0.0], [-0.35, 1.0], 0.0),
        ([-0.35, 1.0], [0.35, 0.0], 0.0),
        ([0.35, 0.0], [0.35, 1.0], 0.0),
    ],
    "O": [
        ([-0.35, 0.35], [0.35, 0.35], -180.0),
        ([-0.35, 0.65], [0.35, 0.65], 180.0),
        ([-0.35, 0.35], [-0.35, 0.65], 0.0),
        ([0.35, 0.35], [0.35, 0.65], 0.0),
    ],
    "P": [
        ([-0.35, 0.0], [-0.35, 1.0], 0.0),
        ([-0.35, 1.0], [0.1, 1.0], 0.0),
        ([-0.35, 0.5], [0.1, 0.5], 0.0),
        ([0.1, 1.0], [0.1, 0.5], 180.0),
    ],
    "Q": [
        ([-0.35, 0.35], [0.35, 0.35], -180.0),
        ([-0.35, 0.65], [0.35, 0.65], 180.0),
        ([-0.35, 0.35], [-0.35, 0.65], 0.0),
        ([0.35, 0.35], [0.35, 0.65], 0.0),
        ([0.0, 0.35], [0.35, 0.0], 0.0),
    ],
    "R": [
        ([-0.35, 0.0], [-0.35, 1.0], 0.0),
        ([-0.35, 1.0], [0.1, 1.0], 0.0),
        ([-0.35, 0.5], [0.1, 0.5], 0.0),
        ([0.1, 1.0], [0.1, 0.5], 180.0),
        ([0.0, 0.5], [0.35, 0.0], 0.0),
    ],
    "S": [
        ([-0.35, 0.0], [0.1, 0.0], 0.0),
        ([0.1, 0.0], [0.1, 0.5], -180.0),
        ([-0.1, 0.5], [0.1, 0.5], 0.0),
        ([-0.1, 0.5], [-0.1, 1.0], 180.0),
        ([-0.1, 1.0], [0.35, 1.0], 0.0),
    ],
    "T": [
        ([0.0, 0.0], [0.0, 1.0], 0.0),
        ([-0.35, 1.0], [0.35, 1.0], 0.0),
    ],
    "U": [
        ([-0.35, 0.35], [0.35, 0.35], -180.0),
        ([-0.35, 0.35], [-0.35, 1.0], 0.0),
        ([0.35, 0.35], [0.35, 1.0], 0.0),
    ],
    "V": [
        ([-0.35, 1.0], [0.0, 0.0], 0.0),
        ([0.35, 1.0], [0.0, 0.0], 0.0),
    ],
    "W": [
        ([-0.35, 1.0], [-0.15, 0.0], 0.0),
        ([-0.15, 0.0], [0.0, 0.65], 0.0),
        ([0.15, 0.0], [0.0, 0.65], 0.0),
        ([0.35, 1.0], [0.15, 0.0], 0.0),
    ],
    "X": [
        ([-0.35, 0.0], [0.35, 1.0], 0.0),
        ([0.35, 0.0], [-0.35, 1.0], 0.0),
    ],
    "Y": [
        ([-0.35, 1.0], [0.0, 0.5], 0.0),
        ([0.35, 1.0], [0.0, 0.5], 0.0),
        ([0.0, 0.5], [0.0, 0.0], 0.0),
    ],
    "Z": [
        ([-0.35, 1.0], [0.35, 1.0], 0.0),
        ([0.35, 1.0], [-0.35, 0.0], 0.0),
        ([-0.35, 0.0], [0.35, 0.0], 0.0),
    ],
    "0": [
        ([-0.15, 0.06], [0.15, 0.06], -105.0),
        ([-0.15, 0.06], [-0.15, 0.94], 60.0),
        ([-0.15, 0.94], [0.15, 0.94], 105.0),
        ([0.15, 0.06], [0.15, 0.94], -60.0),
    ],
    "1": [
        ([0.0, 0.0], [0.0, 1.0], 0.0),
        ([-0.15, 0.0], [0.15, 0.0], 0.0),
        ([-0.2, 0.8], [0.0, 1.0], 0.0),
    ],
    "2": [
        ([-0.25, 0.75], [0.177, 0.573], 225.0),
        ([0.177, 0.573], [-0.25, 0.0], 0.0),
        ([-0.25, 0.0], [0.25, 0.0], 0.0),
    ],
    "3": [
        ([-0.25, 1.0], [0.0, 1.0], 0.0),
        ([0.0, 1.0], [0.0, 0.5], 180.0),
        ([-0.15, 0.5], [0.0, 0.5], 0.0),
        ([0.0, 0.5], [0.0, 0.0], 180.0),
        ([-0.25, 0.0], [0.0, 0.0], 0.0),
    ],
    "4": [
        ([-0.25, 1.0], [-0.25, 0.5], 0.0),
        ([-0.25, 0.5], [0.25, 0.5], 0.0),
        ([0.25, 1.0], [0.25, 0.0], 0.0),
    ],
    "5": [
        ([-0.25, 1.0], [0.25, 1.0], 0.0),
        ([-0.25, 1.0], [-0.25, 0.6], 0.0),
        ([-0.25, 0.6], [0.25, 0.25], 110.0),
        ([0.25, 0.25], [-0.25, 0.25], 180.0),
    ],
    "6": [
        ([0.25, 1.0], [-0.25, 0.25], -66.5),
        ([-0.25, 0.25], [0.25, 0.25], 180.0),
        ([-0.25, 0.25], [0.25, 0.25], -180.0),
    ],
    "7": [
        ([-0.25, 1.0], [0.25, 1.0], 0.0),
        ([0.25, 1.0], [-0.25, 0.0], 0.0),
    ],
    "8": [
        ([-0.25, 0.75], [0.25, 0.75], 180.0),
        ([-0.25, 0.75], [0.25, 0.75], -180.0),
        ([-0.25, 0.25], [0.25, 0.25], 180.0),
        ([-0.25, 0.25], [0.25, 0.25], -180.0),
    ],
    "9": [
        ([-0.25, 0.75], [0.25, 0.75], 180.0),
        ([-0.25, 0.75], [0.25, 0.75], -180.0),
        ([0.25, 0.75], [-0.25, 0.0], 66.5),
    ],
    ".": [([0.0, 0.0], [0.0, 0.0], 0.0)],
    "!": [
        ([0.0, 0.0], [0.0, 0.0], 0.0),
        ([0.0, 1.0], [0.0, 0.25], 0.0),
    ],
    "?": [
        ([0.0, 0.0], [0.0, 0.0], 0.0),
        ([-0.25, 0.75], [0.177, 0.573], 225.0),
        ([0.177, 0.573], [0.0, 0.323], -45.0),
        ([0.0, 0.323], [0.0, 0.25], 0.0),
    ],
    " ": [],
}


class Letter(Lineart2D):
    def __init__(self, char, pos, angle, size, color):
        super(Letter, self).__init__()
        if char in letters:
            segs = letters[char]
            for seg in segs:
                self.add_segment(*seg)
        self.color = color
        self.size = size
        self.off = [0.0, 0.0]
        self.angle = 0.0
        self.speed = (0.0, 0.0)
        self.torque = 0.0
        self.change_pos(pos)
        self.change_angle(angle)

    def change_angle(self, deltaangle):
        self.angle += deltaangle
        ca = m.cos(self.angle * m.pi / 180.0) / self.size
        sa = m.sin(self.angle * m.pi / 180.0) / self.size
        self.mat = [[ca, sa], [-sa, ca]]

    def change_pos(self, delta):
        self.off = [self.off[0] - delta[0], self.off[1] - delta[1]]

    def update(self, step):
        self.change_pos((step * self.speed[0], step * self.speed[1]))
        self.change_angle(step * self.torque)


# Example scenes


class MutatingShapeScene(Scene):
    def __init__(self, ctr):
        super(MutatingShapeScene, self).__init__(ctr)
        self.corners = 4
        self.rot = 0.0
        self.goalrot = random() - 0.5
        self.ratiof = 0.0
        self.goalratiof = random() * 0.8
        self.radius = 0.5
        self.cm = ColorMeander()
        self.shape = Star(
            self.corners, (0, 0), 0.0, self.radius, self.radius, self.cm.get()
        )
        self.add_shape(self.shape)

    def update(self, step):
        # Color
        self.cm.step()
        self.shape.color = self.cm.get()
        # Corners
        if random() < 0.02:
            if random() < (self.corners - 2) / 6.0:
                self.corners -= 1
            else:
                self.corners += 1
            self.shape.num = self.corners
        # Rotation
        if self.goalrot - self.rot > 0.0:
            self.rot += 0.01
            if self.rot > self.goalrot:
                self.goalrot = random() - 0.5
        else:
            self.rot -= 0.01
            if self.rot < self.goalrot:
                self.goalrot = random() - 0.5
        self.shape.set_torque(360.0 / 20.0 * self.rot)
        # Ratio
        if self.goalratiof - self.ratiof > 0.0:
            self.ratiof += 0.01
            if self.ratiof > self.goalratiof:
                self.goalratiof = random() * 0.8
        else:
            self.ratiof -= 0.01
            if self.ratiof < self.goalratiof:
                self.goalratiof = random() * 0.8
        self.shape.set_radius(
            self.radius * (1.0 + self.ratiof), self.radius * (1.0 - self.ratiof)
        )
        # Angle
        self.shape.update(step)


class MovingShapesScene(Scene):
    def __init__(self, ctr):
        super(MovingShapesScene, self).__init__(ctr)
        self.time = 0
        self.crtime = 0
        self.freq = 0.5
        self.horizon = self.preferred_fps * 20
        self.replay = False
        self.record = False
        self.pre_shapes = []
        self.colorfunc = random_hsl_color_func(sat=[0.5,1.0], light=0.0)

    def create(self):
        # slumpa ut form, färg, hastighet, rotation, riktning, offset från centrum, (djup)
        col = self.colorfunc() # hsl_color(random(), 0.6 + 0.4*random(), 0.0)
        speed = m.exp((random() - 0.5) * 1.1) / (self.preferred_fps * 6.0)
        rot = (random() - 0.5) * 360.0 / self.preferred_fps * 0.1
        angle = random() * 2 * m.pi
        offset = (random() - 0.5) * 1.6
        vec = (m.sin(angle), m.cos(angle))
        cent = (-vec[0] * 1.5 + vec[1] * offset, -vec[1] * 1.5 - vec[0] * offset)
        vel = (vec[0] * speed, vec[1] * speed)
        nstep = int(3.0 / speed)
        sp = int(random() * (2 + 4 + 5))
        if sp == 0:  # circle
            sizef = random() * 0.8 + 0.1
            shape = Ellipse(cent, 0.0, sizef*0.5, sizef*0.5, col)
        elif sp == 1:  # ellipse
            ratiof = random() * 0.8
            shape = Ellipse(cent, 0.0, (1.0 + ratiof) * 0.25, (1.0 - ratiof) * 0.25, col)
        elif sp < 6:  # Polygon
            corners = sp + 1
            sizef = random() * 0.8 + 0.1
            shape = Polygon(corners, cent, 0.0, sizef*0.5, col)
        else:  # Star
            corners = sp - 4
            ratiof = random() * 0.8
            shape = Star(
                corners, cent, 0.0, (1.0 + ratiof) * 0.25, (1.0 - ratiof) * 0.25, col
            )
        shape.set_speed(vel[0], vel[1])
        shape.set_torque(rot)
        shape.duetime = self.time + nstep
        return shape

    def reset(self, numframes):
        if numframes and self.horizon:
            self.time = -self.horizon
            self.crtime = self.time
            self.record = True
            while self.time < 0:
                self.update(1)
            self.record = False
        else:
            self.time = 0
            self.crtime = 0
            self.replay = False
            self.record = False

    def update(self, step):
        for sh in reversed(self.shapes):
            if sh.duetime < self.time:
                self.shapes.remove(sh)
        if self.time >= self.crtime:
            if self.record:
                sh = self.create()
                self.add_shape(copy(sh))
                sh.duetime += self.preferred_frames
                self.crtime = self.time + int(-m.log(random()*0.94 + 0.03) / self.freq * self.preferred_fps)
                self.pre_shapes.append((self.crtime + self.preferred_frames, sh))
            elif self.replay:
                if self.pre_shapes:
                    self.add_shape(self.pre_shapes[0][1])
                    self.crtime = self.pre_shapes[0][0]
                    self.pre_shapes = self.pre_shapes[1:]
            else:
                self.add_shape(self.create())
                self.crtime = self.time + int(-m.log(random()*0.94 + 0.03) / self.freq * self.preferred_fps)
                if self.pre_shapes and self.crtime >= self.pre_shapes[0][0]:
                    self.replay = True
        super(MovingShapesScene, self).update(step)
        self.time += step


class RunningText(MovingShapesScene):
    def __init__(self, ctr, txt, color, linewidth=0.1, size=0.6, speed=1.0):
        super(RunningText, self).__init__(ctr)
        self.textcolor = color
        self.lw = linewidth
        self.txt = txt
        self.size = size
        self.speed = speed
        self.horizon = 0
        self.proj2D3D = "cylshell"
        self.place_text()
        self.preferred_frames = self.nsteps

    def set_fps(self, fps):
        self.preferred_fps = fps
        self.place_text()
        self.preferred_frames = self.nsteps
        
    def place_text(self):
        bounds = self.get_scene_bounds()
        # Size is relative the total height of leds, convert to relative radius
        size = self.size * (bounds[1][1] - bounds[1][0])
        # Speed is in letter heights per second, convert to length per step
        speed = size * self.speed / self.preferred_fps
        self.currx = bounds[0][1]
        self.endx = bounds[0][0]
        self.liney = -size / 2.0
        self.shapes = []
        for ind, ch in enumerate(self.txt):
            if isinstance(self.textcolor, list):
                col = self.textcolor[ind % len(self.textcolor)]
            elif callable(self.textcolor):
                col = self.textcolor(ind)
            else:
                col = self.textcolor
            sh = Letter(ch, (0, self.liney), 0, size, col)
            sh.lw = self.lw
            sh.off[0] = -self.currx + (sh.extent[0] - sh.lw * 0.5) * size
            sh.set_speed(-speed, 0.0)
            wdt = max(0.4, sh.extent[2] - sh.extent[0] + 2 * sh.lw)
            self.currx += wdt * size
            self.add_shape(sh)
        self.nsteps = int(round((self.currx - self.endx) / speed + 0.5))
        self.time = 0

    def reset(self, numframes):
        if self.time != 0:
            self.place_text()

    def update(self, step):
        for sh in self.shapes:
            sh.update(step)
        self.time += step


class CaleidoScene(MovingShapesScene):
    def __init__(self, ctr, sym):
        super(CaleidoScene, self).__init__(ctr)
        self.freq = 0.6
        self.symang = m.pi / sym

    def create(self):
        # slumpa ut form, färg, hastighet, rotation, riktning, offset från centrum, (djup)
        col = hsl_color(random(), 1.0, 0.0)
        speed = m.exp((random() - 0.5) * 2.3) / 20.0 * 0.1
        rot = (random() - 0.5) * 360.0 / 20.0 * 0.3
        angle = random() * 2 * m.pi
        offset = random() - 0.5
        vec = (m.sin(angle), m.cos(angle))
        cent = (-vec[0] * 1.5 + vec[1] * offset, -vec[1] * 1.5 - vec[0] * offset + 0.5)
        vel = (vec[0] * speed, vec[1] * speed)
        nstep = int(3.0 / speed)
        sp = int(random() * (2 + 4 + 5))
        if sp == 0:  # circle
            sizef = random() * 0.8 + 0.1
            shape = Ellipse(cent, 0.0, sizef, sizef, col)
        elif sp == 1:  # ellipse
            ratiof = random() * 0.8
            shape = Ellipse(cent, 0.0, (1.0 + ratiof) * 0.5, (1.0 - ratiof) * 0.5, col)
        elif sp < 6:  # Polygon
            corners = sp + 1
            sizef = random() * 0.8 + 0.1
            shape = Polygon(corners, cent, 0.0, sizef, col)
        else:  # Star
            corners = sp - 4
            ratiof = random() * 0.8
            shape = Star(
                corners, cent, 0.0, (1.0 + ratiof) * 0.5, (1.0 - ratiof) * 0.5, col
            )
        shape.set_speed(vel[0], vel[1])
        shape.set_torque(rot)
        shape.duetime = self.time + nstep
        return shape

    def get_color(self, coord, ind):
        r = m.sqrt(coord[0] ** 2 + coord[1] ** 2)
        a = m.atan2(coord[1], coord[0])
        a = abs((a % (2 * self.symang)) - self.symang)
        return super(CaleidoScene, self).get_color((m.sin(a) * r, m.cos(a) * r), ind)


class BouncingScene(MovingShapesScene):
    def __init__(self, ctr, num, speed=0.25, size=0.3, colorfunc=False):
        super(BouncingScene, self).__init__(ctr)
        self.bounds = ctr.get_layout_bounds()
        self.size = size
        self.speed = speed
        self.colfunc = colorfunc or random_hsl_color_func(light=0.0)
        for i in range(num):
            self.ind = i
            self.add_shape(self.create())

    def create(self):
        cent = tuple(uniform(*self.bounds["bounds"][d]) for d in range(self.bounds["dim"]))
        vec = tuple(gauss(0.0, 0.1) for d in range(self.bounds["dim"]))
        vlen = self.dot(vec, vec) ** 0.5
        shape = Blob(cent, self.size/2.0, self.colfunc(self.ind))
        shape.speed = tuple(ele/vlen * self.speed/self.preferred_fps for ele in vec)
        return shape

    def dot(self, v1, v2):
        return sum(map(lambda x1, x2: x1 * x2, v1, v2))

    def scale(self, v1, sc):
        return tuple(map(lambda x1: x1 * sc, v1))

    def add(self, v1, v2):
        return tuple(map(lambda x1, x2: x1 + x2, v1, v2))

    def bounces(self, sh):
        rad = self.dot(sh.cent, sh.cent)
        dir = self.dot(sh.cent, sh.speed)
        if rad > 1.0 and dir > 0.0:
            return True
        for d in range(self.bounds["dim"]):
            c = sh.cent[d] * self.bounds["radius"] + self.bounds["center"][d]
            if ((c < self.bounds["bounds"][d][0] and sh.speed[d] < 0.0) or
                (c > self.bounds["bounds"][d][1] and sh.speed[d] > 0.0)):
                return True
        if self.bounds["dim"] == 3:
            vec = (sh.cent[0], sh.cent[2])
            sp = (sh.speed[0], sh.speed[2])
            rad = self.dot(vec, vec) ** 0.5
            dir = self.dot(vec, sp)
            if rad * self.bounds["radius"] > self.bounds["cylradius"] and dir > 0.0:
                return True
        return False

    def update(self, step):
        for sh in self.shapes:
            if self.bounces(sh):
                vec = tuple(c + gauss(0.0, 0.1) for c in sh.cent)
                vlen2 = self.dot(vec, vec)
                delta = self.dot(vec, sh.speed) / vlen2
                sh.speed = self.add(sh.speed, self.scale(vec, -2 * delta))
            sh.update(step)


# ---------------------------------------------------------




"""
Day 7 - Breathing selected colors

And then an attempt to mimic a pattern from the app again. This is actually
more similar to the Glow effect on the app than the Day 2 example: Each led
stays the same hue but varies in brightness. The hues are picked from a fixed
list. Differences from Glow on the app are: The color positions are random and
not regular, and (just as in the Day 2 example) a smooth transition to the
first frame. 

You can select other colors by changing the cols list below. A second example
is provided, remove the # in front of the line to see it. Colors are given as
(hue, saturation, lightness) where hue and saturation has range [0, 1] and
lightness range [-1, 1] where -1 is black, +1 is white, and 0 the most colored.
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
cols = [(0.375, 1.0, 0.0), (0.58, 1.0, 0.0), (0.375, 1.0, 0.3), (0.29, 1.0, 0.0)]
# cols = [(0.72, 1.0, 0.15), (0.89, 1.0, 0.0), (0.05, 1.0, 0.05), (0.0, 1.0, 0.8)]
BreathCP(ctr, cols).launch_movie()



"""
xled_plus.effects
~~~~~~~~~~~~

Author: Anders Holst (anders.holst@ri.se), 2021

A collection of moving effects created as subclasses of the Effect base class.
Mostly layout-independent effects, i.e with randomized positions, such as
glowing, breathing, and sparkling.
Most effects are highly configurable. Some specific examples of each effect
type are also provided.
"""

from xled_plus.effect_base import Effect
from xled_plus.colormeander import ColorMeander
from xled_plus.pattern import (
    blendcolors,
    dimcolor,
    randompoisson,
    randomdiscrete,
    random_hsl_color_func,
    sprinkle_pattern,
)
from xled_plus.ledcolor import hsl_color
import random


"""
Glowing effect

Similar to the Glow effect in the app, but seamless when it wraps around.
Check out the specific examples: Charcoal, Fire, Water, Aurora, Meadow.
"""


class Glowbit:
    def __init__(self, cols, bend, steps, initstep=False, loop=False):
        self.count = 0
        self.lastcol = (0, 0, 0)
        self.nextcol = (0, 0, 0)
        self.currcol = (0, 0, 0)
        self.loop = loop
        self.currind = initstep if initstep else steps
        self.steps = steps
        self.cols = cols
        self.bend = bend
        if self.loop:
            self.initcol1 = hsl_color(
                *self.cols[int((random.random() ** self.bend) * len(self.cols))]
            )
            self.initcol2 = hsl_color(
                *self.cols[int((random.random() ** self.bend) * len(self.cols))]
            )
            self.lastcol = self.initcol1
            self.nextcol = self.initcol2

    def getnext(self):
        if self.currind == self.steps:
            self.lastcol = self.nextcol
            if self.loop and self.count + self.steps >= self.loop:
                self.nextcol = self.initcol2
            elif self.loop and self.count + 2 * self.steps >= self.loop:
                self.nextcol = self.initcol1
            else:
                self.nextcol = hsl_color(
                    *self.cols[int((random.random() ** self.bend) * len(self.cols))]
                )
            self.currind = 0
        self.currind += 1
        self.count += 1
        return blendcolors(self.lastcol, self.nextcol, float(self.currind) / self.steps)


class GlowEffect(Effect):
    def __init__(self, ctr, cols, bend, cycles, fps=False):
        super(GlowEffect, self).__init__(ctr)
        if fps:
            self.preferred_fps = fps
        self.cols = cols
        self.bend = bend
        self.cycles = cycles

    def reset(self, numframes):
        if type(self.cycles) == int:
            steps = [self.cycles]
        elif numframes:
            steps = []
            for n in range(self.cycles[0], self.cycles[-1] + 1):
                if numframes % n == 0:
                    steps.append(n)
            if not steps:
                steps = list(range(self.cycles[0], self.cycles[-1] + 1))
        else:
            steps = list(range(self.cycles[0], self.cycles[-1] + 1))
        pr1 = 13 if len(steps) % 13 != 0 else 7
        pr2 = 11 if len(steps) % 11 != 0 else 7
        self.glowarray = [
            Glowbit(
                self.cols,
                self.bend,
                steps[(i * pr1) % len(steps)],
                (i * pr2) % steps[(i * pr1) % len(steps)],
                numframes,
            )
            for i in range(self.ctr.num_leds)
        ]

    def getnext(self):
        return self.ctr.make_func_pattern(lambda i: self.glowarray[i].getnext())


class Charcoal(GlowEffect):
    def __init__(self, ctr):
        cols = [
            [0.6057, 1.0, -0.99],
            [0.6171, 1.0, -0.8865],
            [0.6254, 1.0, -0.8083],
            [0.5796, 1.0, -0.7257],
            [0.6222, 1.0, -0.6779],
            [0.5606, 1.0, -0.6080],
            [0.5956, 1.0, -0.4838],
        ]
        super(Charcoal, self).__init__(ctr, cols, 2, [2, 4], 8)


class Fire(GlowEffect):
    def __init__(self, ctr):
        cols = [
            [0.5689, 1.0, -0.2847],
            [0.5413, 1.0, -0.1809],
            [0.5119, 1.0, -0.0685],
            [0.6185, 1.0, -0.4416],
            [0.6206, 1.0, -0.6780],
            [0.5068, 1.0, 0.1797],
            [0.5603, 1.0, -0.0170],
            [0.45, 1.0, 0.1],
        ]
        super(Fire, self).__init__(ctr, cols, 2, [3, 6], 20)


class Water(GlowEffect):
    def __init__(self, ctr):
        cols = [
            [0.0, 1.0, -0.2],
            [0.0, 1.0, -0.5],
            [0.05, 1.0, -0.3],
            [0.05, 1.0, 0.0],
            [0.1, 1.0, -0.5],
            [0.1, 1.0, 0.0],
            [0.15, 1.0, 0.0],
            [0.0, 1.0, 0.8],
        ]
        super(Water, self).__init__(ctr, cols, 2, [3, 6], 20)


class Meadow(GlowEffect):
    def __init__(self, ctr):
        cols = [
            [0.2427, 1.0, -0.6294],
            [0.2556, 1.0, -0.3245],
            [0.2692, 1.0, -0.0834],
            [0.2456, 1.0, 0.0243],
            [0.2901, 1.0, 0.2506],
            [0.400, 1.0, 0.4219],
            [0.6065, 1.0, -0.1989],
            [0.7709, 1.0, -0.3420],
            [0.7833, 0.1259, 0.1001],
        ]
        super(Meadow, self).__init__(ctr, cols, 2, [4, 8])


class Aurora(GlowEffect):
    def __init__(self, ctr):
        cols = [
            [0.0, 0.0, -0.8],
            [0.0, 0.0, -0.5],
            [0.0, 0.0, 0.0],
            [0.7401, 0.3679, 0.1246],
            [0.2744, 0.3180, 0.1759],
            [0.2789, 0.7024, 0.2672],
            [0.7435, 0.7191, 0.2302],
            [0.7483, 1.0, 0.1677],
            [0.1928, 1.0, -0.1506],
            [0.1, 1.0, 0.9],
        ]
        super(Aurora, self).__init__(ctr, cols, 2, [6, 10])


class Brown(GlowEffect):
    def __init__(self, ctr):
        cols = [
            (0.5, 0.0, -1.0),
            (0.435, 1.0, -0.98),
            (0.477, 1.0, -0.96),
            (0.497, 1.0, -0.81),
            (0.45, 1.0, -0.765),
            (0.52, 1.0, -0.76),
        ]
        super(Brown, self).__init__(ctr, cols, 3, [6, 10])


class GlowCP(GlowEffect):
    def __init__(self, ctr, cols):
        super(GlowCP, self).__init__(ctr, cols, 3, [4, 8])


"""
Sparkling effect

Similar to the Bright Twinkle effect in the app, but more versatile.
Again there are specific examples, but check out SparkleStars which has
varied color temperature of whites, unlike the origonal.
"""


class SparkleEffect(Effect):
    def __init__(self, ctr, freq, nfunc, sfunc, icol=(0, 0, 0)):
        super(SparkleEffect, self).__init__(ctr)
        self.freq = freq
        self.newfunc = nfunc
        self.stepfunc = sfunc
        self.initialcol = icol

    def reset(self, numframes):
        self.pattern = self.ctr.make_solid_pattern(self.initialcol)
        self.time = -1
        self.slist = []
        self.olist = list(range(0, self.ctr.num_leds))
        self.numframes = False  # intentionally set to False here
        if numframes:
            # walk a number of steps, count the sfunc cycle, and record the poisson outcomes
            self.blocktime = {}
            tmp = self.newfunc(0, 0)
            while self.stepfunc(0, self.time + 1, tmp) not in [False, True]:
                self.getnext()
            self.leadintime = self.time
            self.leadin = [ele for ele in self.slist]
            self.numframes = numframes

    def getnext(self):
        self.time += 1
        if self.numframes:
            self.pattern = self.ctr.copy_pattern(self.pattern)  # only needed for movie
        if self.numframes and self.time >= self.numframes:
            for (ind, coldesc, tm) in self.leadin:
                if tm == self.time - self.numframes:
                    self.slist.append((ind, coldesc, self.time))
        else:
            if self.numframes and self.time >= self.numframes - self.leadintime:
                for (ind, coldesc, tm) in self.leadin:
                    if (
                        tm <= self.time - self.numframes + self.leadintime
                        and ind in self.olist
                    ):
                        self.olist.remove(ind)
            n = randompoisson(self.freq)
            for j in range(n):
                if self.olist:
                    pos = random.randint(0, len(self.olist) - 1)
                    ind = self.olist[pos]
                    coldesc = self.newfunc(ind, self.time)
                    self.slist.append((ind, coldesc, self.time))
                    del self.olist[pos]
        remlst = []
        for pos, (ind, coldesc, stime) in enumerate(self.slist):
            col = self.stepfunc(ind, self.time - stime, coldesc)
            if col is False:
                remlst.append(pos)
                self.olist.append(ind)
            elif col is True:
                remlst.append(pos)
                self.olist.append(ind)
                self.ctr.modify_pattern(self.pattern, ind, self.initialcol)
            else:
                self.ctr.modify_pattern(self.pattern, ind, col)
        for pos in reversed(sorted(remlst)):
            del self.slist[pos]
        return self.pattern


def random_color_func(hue=False, sat=False, light=False):
    return random_hsl_color_func(hue, sat, light)


def random_hs_func(hue=False, sat=False):
    def isnum(x):
        return type(x) in [float, int]

    if isnum(hue):
        hue = [hue, 0.0]
    elif hue is False:
        hue = [0.0, 1.0]
    if isnum(sat):
        sat = [sat, 0.0]
    elif hue is False:
        sat = [0.0, 1.0]

    def func(*args):
        h = (
            ((hue[1] - hue[0]) * random.random() + hue[0]) % 1.0
            if hue[1] > 0
            else hue[0]
        )
        s = ((sat[1] - sat[0]) * random.random() + sat[0]) if sat[1] > 0 else sat[0]
        return (h, s)

    return func


def circular_hs_func(cycle, dists=False, probs=False):
    def func0(ind, tm):
        return ((tm / float(cycle)) % 1.0, 1.0)

    def func1(ind, tm):
        return ((tm / float(cycle) + dists[randomdiscrete(probs)]) % 1.0, 1.0)

    def func2(ind, tm):
        return ((tm / float(cycle) + random.choice(dists)) % 1.0, 1.0)

    if not dists:
        return func0
    elif probs and len(probs) == len(dists):
        return func1
    else:
        return func2


def circular_color_func(cycle, dists=False, probs=False, light=0.0):
    func = circular_hs_func(cycle, dists, probs)
    return lambda ind, tm: hsl_color(*func(ind, tm), light)


def tinted_white_func(hue1, depth1, hue2, depth2):
    sc = depth1 + depth2
    mid = depth1 / sc

    def func(*args):
        r = random.random()
        return hsl_color(hue1 if r < mid else hue2, 1.0, 1.0 - sc * abs(r - mid))

    return func


def selected_color_func(cols, probs=False):
    def func1(ind, tm):
        return cols[randomdiscrete(probs)]

    def func2(ind, tm):
        return random.choice(cols)

    if probs and len(probs) == len(cols):
        return func1
    else:
        return func2


def pulselight_func(up, stable, down, initcol=(0, 0, 0), lin=False):
    tot = up + stable + down

    def func(ind, tm, rgb):
        if tm < up:
            pr = (tm + 1.0) / (up + 1.0)
            return blendcolors(initcol, rgb, pr if lin else pr * pr)
        elif tm < up + stable:
            return rgb
        elif tm < tot:
            pr = (tot - tm) / (down + 1.0)
            return blendcolors(initcol, rgb, pr if lin else pr * pr)
        else:
            return True

    return func


def flashlight_func(stable, down, flashcol=False, initcol=(0, 0, 0), lin=False):
    if not flashcol:
        flashcol = hsl_color(0.0, 0.0, 1.0)
    tot = 1 + stable + down

    def func(ind, tm, rgb):
        if tm < 1:
            return flashcol
        elif tm < 1 + stable:
            return rgb
        elif tm < tot:
            pr = (tot - tm) / (down + 1.0)
            return blendcolors(initcol, rgb, pr if lin else pr * pr)
        else:
            return True

    return func


def looplight_func(up, down, sprop_up=1.0, sprop_down=0.0, lin=False):
    tot = up + down

    def func(ind, tm, hs):
        if tm < up:
            pr = (tm + 1.0) / up
            if not lin:
                pr *= pr
            return hsl_color(hs[0], hs[1] * sprop_up, 2 * pr - 1.0)
        elif tm < tot:
            pr = (tot - tm) / down
            if not lin:
                pr *= pr
            return hsl_color(hs[0], hs[1] * sprop_down, 2 * pr - 1.0)
        else:
            return True

    return func


class SimpleBlink(SparkleEffect):
    def __init__(self, ctr):
        super(SimpleBlink, self).__init__(
            ctr, 8, random_color_func(sat=1.0, light=0.0), pulselight_func(0, 1, 0)
        )


class SimpleStars(SparkleEffect):
    def __init__(self, ctr):
        white = hsl_color(0.0, 0.0, 1.0)
        super(SimpleStars, self).__init__(
            ctr, 8, lambda *args: white, pulselight_func(0, 1, 0)
        )


class Pulselight(SparkleEffect):
    def __init__(self, ctr):
        super(Pulselight, self).__init__(
            ctr, 3, random_color_func(light=0.0), pulselight_func(18, 4, 18)
        )
        self.preferred_fps = 12


class Looplight(SparkleEffect):
    def __init__(self, ctr, reverse=False):
        if reverse:
            sfunc = looplight_func(8, 16, sprop_up=0.0, sprop_down=1.0)
        else:
            sfunc = looplight_func(16, 8)
        super(Looplight, self).__init__(ctr, 4, random_hs_func(sat=[0.5, 1.0]), sfunc)


class LooplightSpectrum(SparkleEffect):
    def __init__(self, ctr, cycle=240):
        super(LooplightSpectrum, self).__init__(
            ctr, 4, circular_hs_func(cycle), looplight_func(16, 8)
        )
        self.preferred_frames = cycle


class SparkleRandom(SparkleEffect):
    def __init__(self, ctr, hue=False, sat=False, light=False):
        super(SparkleRandom, self).__init__(
            ctr, 3, random_color_func(hue, sat, light), pulselight_func(16, 8, 16)
        )
        self.preferred_fps = 12


class SparkleStars(SparkleEffect):
    def __init__(self, ctr):
        super(SparkleStars, self).__init__(
            ctr, 3, tinted_white_func(0.0, 0.5, 0.5, 0.5), pulselight_func(16, 8, 16)
        )
        self.preferred_fps = 12


class SparkleCP(SparkleEffect):
    def __init__(self, ctr, cols):
        colsrgb = list(map(lambda hsl: hsl_color(*hsl), cols))
        super(SparkleCP, self).__init__(
            ctr, 3, selected_color_func(colsrgb), pulselight_func(16, 8, 16)
        )
        self.preferred_fps = 12


"""
Breathing effect

Each led has a fixed color but slowly pulsing brightness.
"""


class Breathbit:
    def __init__(self, col, lspan, steps, stayhigh=0, staylow=0, initstep=False):
        self.currind = initstep if initstep else steps
        self.steps = steps
        self.hsteps = (steps - 1) / 2.0
        self.lspan = 1.0 - (1.0 - lspan) ** 0.5
        self.setcolor(col)
        self.preferred_fps = 6
        self.preferred_frames = 60

    def setcolor(self, col):
        self.col = hsl_color(*col)

    def getnext(self):
        self.currind += 1
        if self.currind >= self.steps:
            self.currind = 0
        prop = abs(self.currind - self.hsteps) / self.hsteps
        return dimcolor(self.col, (prop * self.lspan + 1.0 - self.lspan) ** 2)


class BreathEffect(Effect):
    def __init__(self, ctr, cols, bend, lspan, cycles, fps=False):
        super(BreathEffect, self).__init__(ctr)
        if fps:
            self.preferred_fps = fps
        self.cols = cols
        self.bend = bend
        self.lspan = lspan
        self.cycles = cycles

    def reset(self, numframes):
        if type(self.cycles) == int:
            steps = [self.cycles]
        elif numframes:
            steps = []
            for n in range(self.cycles[0], self.cycles[-1] + 1):
                if numframes % n == 0:
                    steps.append(n)
            if not steps:
                steps = list(range(self.cycles[0], self.cycles[-1] + 1))
        else:
            steps = list(range(self.cycles[0], self.cycles[-1] + 1))
        pr1 = 13 if len(steps) % 13 != 0 else 7
        pr2 = 11 if len(steps) % 11 != 0 else 7
        colarray = [
            self.cols[int((random.random() ** self.bend) * len(self.cols))]
            for i in range(self.ctr.num_leds)
        ]
        self.brarray = [
            Breathbit(
                colarray[i],
                self.lspan,
                steps[(i * pr1) % len(steps)],
                initstep=(i * pr2) % steps[(i * pr1) % len(steps)],
            )
            for i in range(self.ctr.num_leds)
        ]

    def getnext(self):
        return self.ctr.make_func_pattern(lambda i: self.brarray[i].getnext())


class BreathCP(BreathEffect):
    def __init__(self, ctr, cols):
        super(BreathCP, self).__init__(ctr, cols, 1, 0.75, [12, 30])


"""
Glitter effect

Similar to the Sparkles effect in the app, of brigh flashing leds against a
solid background.
Specific examples are included in an attempt to mimic the "metallic luster"
effects of AWW leds: Gold, Silver, Bronze, RoseGold
"""


class GlitterEffect(Effect):
    def __init__(self, ctr, freq, cols, icol=(0, 0, 0)):
        super(GlitterEffect, self).__init__(ctr)
        self.freq = freq
        self.cols = list(map(lambda hsl: hsl_color(*hsl), cols))
        self.initialcol = hsl_color(*icol)
        self.preferred_fps = 10
        self.preferred_frames = 100

    def reset(self, numframes):
        self.pattern = self.ctr.make_solid_pattern(self.initialcol)

    def getnext(self):
        return sprinkle_pattern(self.ctr, self.pattern, self.cols, self.freq)


class Silver(GlitterEffect):
    def __init__(self, ctr):
        super(Silver, self).__init__(ctr, 10, [(0.0, 1.0, 0.8)], (0.0, 1.0, 0.3))


class Gold(GlitterEffect):
    def __init__(self, ctr):
        super(Gold, self).__init__(ctr, 10, [(0.45, 1.0, 0.5)], (0.45, 1.0, -0.1))


class RoseGold(GlitterEffect):
    def __init__(self, ctr):
        super(RoseGold, self).__init__(ctr, 10, [(0.53, 1.0, 0.6)], (0.53, 1.0, 0.1))


class Bronze(GlitterEffect):
    def __init__(self, ctr):
        super(Bronze, self).__init__(ctr, 10, [(0.58, 1.0, 0.4)], (0.58, 1.0, -0.4))


class StainlessSteel(GlitterEffect):
    def __init__(self, ctr):
        super(StainlessSteel, self).__init__(
            ctr, 10, [(0.0, 0.0, 1.0)], (0.0, 0.0, 0.0)
        )


class GlitterCP(GlitterEffect):
    def __init__(self, ctr, cols):
        if len(cols) == 1:
            lcol = (cols[0][0], cols[0][1], min(1.0, cols[0][2] + 0.5))
            cols.append(lcol)
        super(GlitterCP, self).__init__(ctr, 10, cols[1:], cols[0])


"""
Rotating pattern effect

Each led passes through a sequence by "rotating" the entire pattern.
Two scattered spectrum effects are included as examples, where each led
rapidly passes through the spectrum wheras the overall impression stays
constant.
"""


class RotateEffect(Effect):
    def __init__(self, ctr, pat, perm, step=1, speed=20):
        super(RotateEffect, self).__init__(ctr)
        self.origpattern = pat
        self.perm = perm
        self.step = step
        self.preferred_frames = ctr.num_leds // step
        self.preferred_fps = speed

    def reset(self, numframes):
        self.pattern = self.ctr.copy_pattern(self.origpattern)

    def getnext(self):
        currpattern = (
            self.ctr.permute_pattern(self.pattern, self.perm, circular=True)
            if self.perm
            else self.pattern
        )
        self.pattern = self.ctr.rotate_pattern(self.pattern, self.step, circular=True)
        return currpattern


class Spectrum(RotateEffect):
    def __init__(self, ctr, scattered=False, lightness=0.0, step=1):
        numleds = ctr.num_leds
        pat = ctr.make_func_pattern(
            lambda i: hsl_color(i / float(numleds), 1.0, lightness), circular=True
        )
        if scattered:
            perm = list(range(numleds))
            random.shuffle(perm)
        else:
            perm = False  # [i if i<numleds/2 else (numleds*3)//2 - 1 - i for i in range(numleds)]
        super(Spectrum, self).__init__(ctr, pat, perm, step=step)


class ScatteredSpectrum(Spectrum):
    def __init__(self, ctr):
        super(ScatteredSpectrum, self).__init__(ctr, True, 0.0)


class RotatingWhites(Spectrum):
    def __init__(self, ctr):
        super(RotatingWhites, self).__init__(ctr, True, 0.7, 5)


"""
Color meandering effect

Slowly and randomly changing color, such that whenever you look you will
never see it change, but whenever you look back after a while it has a
different color than last. Is not suitable for a movie, but requires
constant real time feeding, since it quickly consumes the available movie
capacity and never repeats exactly the same sequence of colors again.

Has a few different styles:
'solid' - all leds have the same color at every instance
'sequence' - the color changes propagate along the string
'scattered' - as sequence but with random permutation of the ordering of leds
'multi' - gives three random sequneces at once (every third led being the same color)
'tandem' - first half of string has one color and the second half the complement color
'blend' - two random sequences and each led has a color which is a random blend between them
"""


class ColorMeanderEffect(Effect):
    def __init__(self, ctr, style):
        super(ColorMeanderEffect, self).__init__(ctr)
        # styles: solid, sequence, scattered, multi, blend, tandem, (whites?)
        self.style = style
        self.pat = None
        self.cm = None
        self.updatefunc = None
        self.preferred_fps = 2
        self.preferred_frames = 500

    def reset(self, numframes):
        self.pat = self.ctr.make_func_pattern(lambda i: hsl_color(0, 0, 1))
        self.cm = ColorMeander()
        if numframes:
            self.cm.steplen *= 10
            self.cm.noiselev *= 3
            self.preferred_fps /= 10
        if self.style == "solid":
            self.updatefunc = self.update_solid
        elif self.style == "sequence":
            self.updatefunc = self.update_sequence
        elif self.style == "scattered":
            self.updatefunc = self.update_scattered
            self.perm = list(range(self.ctr.num_leds))
            random.shuffle(self.perm)
            self.pat0 = self.pat
        elif self.style == "tandem":
            self.updatefunc = self.update_tandem
        elif self.style == "multi":
            self.cms = [ColorMeander() for i in range(3)]
            self.updatefunc = self.update_multi
        elif self.style == "blend":
            self.cm2 = ColorMeander()
            self.props = [random.random() for i in range(self.ctr.num_leds)]
            self.updatefunc = self.update_blend
        else:
            print("Bad Meander style")
            self.update_func = lambda: None

    def update_solid(self):
        self.cm.step()
        self.pat = self.ctr.make_func_pattern(lambda i: self.cm.get())

    def update_sequence(self):
        self.cm.step()
        self.pat = self.ctr.shift_pattern(self.pat, 1, self.cm.get(), circular=True)

    def update_scattered(self):
        self.cm.step()
        self.pat0 = self.ctr.shift_pattern(self.pat0, 1, self.cm.get())
        self.pat = self.ctr.permute_pattern(self.pat0, self.perm)

    def update_tandem(self):
        self.cm.step()
        (h, s, l) = self.cm.get_hsl()
        col1 = hsl_color(h, s, l)
        col2 = hsl_color((h + 0.5) % 1.0, s, l)
        self.pat = self.ctr.make_func_pattern(
            lambda i: col1 if i < self.ctr.num_leds // 2 else col2
        )

    def update_multi(self):
        for cm in self.cms:
            cm.step()
        self.pat = self.ctr.make_func_pattern(
            lambda i: self.cms[i % len(self.cms)].get()
        )

    def update_blend(self):
        self.cm.step()
        self.cm2.step()
        self.pat = self.ctr.make_func_pattern(
            lambda i: blendcolors(self.cm.get(), self.cm2.get(), self.props[i])
        )

    def getnext(self):
        self.updatefunc()
        return self.pat
